import numpy as np
import pyvista as pv
import scipy.ndimage as ndi
import matplotlib.pyplot as plt


import vreg.utils
import vreg.mod_affine
import vreg.vol as vol


def _mosaic(volume:np.ndarray):

    if isinstance(volume, vol.Volume3D):
        vals = volume.values
        n = vals.shape[2]
        nrows = np.ceil(np.sqrt(n)).astype(int)
        ncols = np.ceil(n/nrows).astype(int)
        nx = vals.shape[0]
        ny = vals.shape[1]
        mosaic = np.zeros((nx*ncols, ny*nrows))
        k=0
        for j in range(nrows):
            for i in range(ncols):
                if k < n:
                    mosaic[i*nx:(i+1)*nx, 
                           j*ny:(j+1)*ny] = vals[:,:,k]
                k += 1
        return mosaic

    elif isinstance(volume, list):
        n = len(volume)
        nrows = np.ceil(np.sqrt(n)).astype(int)
        ncols = np.ceil(n/nrows).astype(int)
        nx = volume[0].shape[0]
        ny = volume[0].shape[1]
        mosaic = np.zeros((nx*ncols, ny*nrows))
        k=0
        for j in range(nrows):
            for i in range(ncols):
                if k < n:
                    mosaic[i*nx:(i+1)*nx, 
                           j*ny:(j+1)*ny] = volume[k].values[:,:,0]
                k += 1
        return mosaic


def overlay_2d(under:vol.Volume3D | list, over=None, 
               alpha=0.25, title=None, vmin=None, vmax=None, width=3, dpi=300):
    """Show a volume in 2D with a transparent overlay.

    Args:
        under (vol.Volume3D | list): volume to use as underlay. This can also be 
          a multi-slice volume entered as a list of 2D volumes.
        over (vol.Volume3D, optional): transparent overlay volume.
        alpha (float, optional): transparancy. Defaults to 0.25.
        title (str, optional): Plot title. Defaults to None.
        vmin (float, optional): minimal window value. Defaults to None.
        vmax (float, optional): maximal window value. Defaults to None.
        size (float, optional): Figure width in inches. Defaults to 3.
        dpi (int, optional): Dots per inch. Defaults to 300.

    Raises:
        ValueError: if the arguments are of incorrect type.
    """

    uim = _mosaic(under)
    height = width*uim.shape[1]/uim.shape[0]
    fig, ax = plt.subplots(1, 1, figsize=(width, height), dpi=dpi)

    if vmin is None:
        vmin = np.percentile(uim, 1)
    if vmax is None:
        vmax = np.percentile(uim, 99)

    if over is None:
        _plot_2d(ax, uim, vmin=vmin, vmax=vmax)
    else:
        if isinstance(under, list):
            over = [over.slice_like(u) for u in under]
        else:
            over = over.slice_like(under)
        oim = _mosaic(over)
        _plot_overlay_2d(ax, uim, oim, alpha, vmin=vmin, vmax=vmax)

    plt.show()



def _plot_2d(ax, under, vmin=None, vmax=None):

    # Show the first image
    ax.imshow(under.T, cmap='gray', vmin=vmin, vmax=vmax) 

    # Optional: Add a colorbar
    #plt.colorbar(ax.imshow(image2, cmap='jet', alpha=0.5), ax=ax)
    ax.axis('off')
    # ax.set_xticks([])  
    # ax.set_yticks([]) 


def _plot_overlay_2d(ax, under, over, alpha=0.25, vmin=None, vmax=None):

    # Show the first image
    ax.imshow(under.T, cmap='gray', alpha=1.0, vmin=vmin, vmax=vmax)  # Adjust alpha for transparency

    # Create a mask for image2 where values are non-zero
    mask = over > 0

    # Show the second image only where the mask is True, with full opacity for non-zero values
    # Use np.where to set the alpha values based on the mask
    alpha_im = np.zeros_like(over)  # Create an alpha channel for image2
    alpha_im[mask] = alpha  # Set alpha to 1 for non-zero values

    # Display image2 with the mask applied
    # Using 'jet' colormap for image2, but only where mask is True
    ax.imshow(over.T, cmap='jet', alpha=alpha_im.T)  # Non-zero values will be opaque

    # Optional: Add a colorbar
    #plt.colorbar(ax.imshow(image2, cmap='jet', alpha=0.5), ax=ax)
    ax.axis('off')
    # ax.set_xticks([])  
    # ax.set_yticks([]) 


def pv_contour(values, data, affine, surface=False):

    # For display of the surface, interpolate from volume to surface array
    surf_shape = 1 + np.array(data.shape)
    r = vreg.utils.surface_coordinates(data.shape)
    r = vreg.utils.extend_border(r, data.shape)
    surf_data = ndi.map_coordinates(data, r.T, order=3)
    surf_data = np.reshape(surf_data, surf_shape)

    rotation, translation, pixel_spacing = vreg.utils.affine_components(affine)
    grid = pv.ImageData(dimensions=surf_shape, spacing=pixel_spacing)
    surf = grid.contour(values, surf_data.flatten(order="F"), method='marching_cubes')
    surf = surf.rotate_vector(rotation, np.linalg.norm(rotation)*180/np.pi, inplace=False)
    surf = surf.translate(translation, inplace=False)
    if surface:
        surf = surf.reconstruct_surface()
    return surf


def plot_volume(volume, affine, off_screen=False):
    """Plot

    Args:
        volume (_type_): _description_
        affine (_type_): _description_
        off_screen (bool, optional): _description_. Defaults to False.

    Returns:
        pyvista.Plotter: _description_

    Example:

    .. pyvista-plot::
        :caption: A volume
        :include-source: True

        >>> import vreg
        >>> arr, aff = vreg.generate_plot_data_1()
        >>> pl = vreg.plot_volume(arr, aff, off_screen=True)
        >>> out = pl.show()
    """

    clr, op = (255,255,255), 0.5
    #clr, op = (255,0,0), 1.0

    pl = pv.Plotter(off_screen=off_screen)
    pl.add_axes()

    # Plot the surface
    surf = pv_contour([0.5], volume, affine)
    if len(surf.points) == 0:
        print('Cannot plot the reference surface. It has no points inside the volume. ')
    else:
        pl.add_mesh(surf,
            color=clr, 
            opacity=op,
            show_edges=False, 
            smooth_shading=True, 
            specular=0, 
            show_scalar_bar=False,        
        )

    # Plot wireframe around edges of reference volume
    vertices, faces = vreg.utils.parallellepid(volume.shape, affine=affine)
    box = pv.PolyData(vertices, faces)
    pl.add_mesh(box,
        style='wireframe',
        opacity=op,
        color=clr, 
    )
    
    return pl


def plot_affine_resliced(volume, affine, volume_resliced, affine_resliced, off_screen=False):

    clr, op = (255,0,0), 1.0

    pl = plot_volume(volume, affine, off_screen=off_screen)

    # Plot the resliced surface
    surf = pv_contour([0.5], volume_resliced, affine_resliced)
    if len(surf.points) == 0:
        print('Cannot plot the resliced surface. It has no points inside the volume. ')
    else:
        pl.add_mesh(surf,
            color=clr, 
            opacity=op,
            show_edges=False, 
            smooth_shading=True, 
            specular=0, 
            show_scalar_bar=False,        
        )

    # Plot wireframe around edges of resliced volume
    vertices, faces = vreg.utils.parallellepid(volume_resliced.shape, affine=affine_resliced)
    box = pv.PolyData(vertices, faces)
    pl.add_mesh(box,
        style='wireframe',
        opacity=op,
        color=clr, 
    ) 

    return pl


def plot_affine_transformed(input_data, input_affine, output_data, output_affine, transformation, off_screen=False):

    pl = plot_affine_resliced(input_data, input_affine, output_data, output_affine, off_screen=off_screen)

    # Plot the reference surface
    surf = pv_contour([0.5], output_data, output_affine)
    if len(surf.points) == 0:
        print('Cannot plot the surface. It has no points inside the volume. ')
    else:
        pl.add_mesh(surf,
            color=(0,0,255), 
            opacity=0.25,
            show_edges=False, 
            smooth_shading=True, 
            specular=0, 
            show_scalar_bar=False,        
        )
    
    # Create blue reference box showing transformation
    vertices, faces = vreg.utils.parallellepid(input_data.shape, affine=np.dot(transformation, input_affine)) 
    box = pv.PolyData(vertices, faces)
    pl.add_mesh(box,
        style='wireframe',
        color=(0,0,255),
        opacity=0.5,
    )            
        
    pl.show()


def plot_bounding_box(input_data, input_affine, output_shape, output_affine, off_screen=False):

    pl = plot_volume(input_data, input_affine, off_screen=off_screen)

    # Plot wireframe around edges of resliced volume
    vertices, faces = vreg.utils.parallellepid(output_shape, affine=output_affine)
    box = pv.PolyData(vertices, faces)
    pl.add_mesh(box,
        style='wireframe',
        opacity=1.0,
        color=(255,0,0), 
    )
    pl.show()


def plot_freeform_transformed(input_data, input_affine, output_data, output_affine, off_screen=False):

    pl = plot_affine_resliced(input_data, input_affine, output_data, output_affine, off_screen=off_screen)

    # Plot the reference surface
    surf = pv_contour([0.5], output_data, output_affine)
    if len(surf.points) == 0:
        print('Cannot plot the surface. It has no points inside the volume. ')
    else:
        pl.add_mesh(surf,
            color=(0,0,255), 
            opacity=0.25,
            show_edges=False, 
            smooth_shading=True, 
            specular=0, 
            show_scalar_bar=False,        
        )           
        
    pl.show()


def plot_affine_transform_reslice(input_data, input_affine, output_data, output_affine, transformation, off_screen=False):
    """Visualise the result of an affine transformation with reslicing.

    The original volume is shown in grey, and the resulting volume in red. In 
    blue is shown the transformed volume before reslicing to the final shape.

    Args:
        input_data (numpy.ndarray): Volume3D array before transformation.
        input_affine (numpy.ndarray): Volume3D affine before transformation.
        output_data (numpy.ndarray): Volume3D array after transformation.
        output_affine (numpy.ndarray): Volume3D affine after transformation.
        transformation (numpy.ndarray): affine matrix of the transformation.
        off_screen (bool, optional): Show the result (True) or not (False). 
          Defaults to False.

    Example:

    .. pyvista-plot::
        :caption: Active affine transform with reslicing.
        :include-source: True

        >>> import vreg
        >>> iarr, iaff, oarr, oaff, transl = vreg.generate_translated_data_2()
        >>> transfo = vreg.affine_matrix(translation=transl)
        >>> vreg.plot_affine_transform_reslice(iarr, iaff, oarr, oaff, transfo, off_screen=True)
    """

    # Plot reslice
    pl = plot_affine_resliced(input_data, input_affine, output_data, output_affine, off_screen=off_screen)
    
    # Show in blue transparent the transformation without reslicing
    output_data, output_affine = vreg.mod_affine.affine_transform(input_data, input_affine, transformation, reshape=True)

    # Plot the reference surface
    surf = pv_contour([0.5], output_data, output_affine)
    if len(surf.points) == 0:
        print('Cannot plot the surface. It has no points inside the volume. ')
    else:
        pl.add_mesh(surf,
            color=(0,0,255), 
            opacity=0.25,
            show_edges=False, 
            smooth_shading=True, 
            specular=0, 
            show_scalar_bar=False,        
        )

    # Create blue reference box showing transformation
    vertices, faces = vreg.utils.parallellepid(input_data.shape, affine=np.dot(transformation, input_affine))
    box = pv.PolyData(vertices, faces)
    pl.add_mesh(box,
        style='wireframe',
        color=(0,0,255),
        opacity=0.5,
    )  

    pl.show()


import numpy as np

try:
    import nibabel as nib
except ImportError:
    nib_installed = False
else:
    nib_installed = True

try:
    import zarr
except ImportError:
    zarr_installed = False
else:
    zarr_installed = True

from vreg.vol import Volume3D


NIFTI_IMPORT_ERROR = (
    "Saving in NIfTI format requires the nibabel python package. You can "
    "install it with 'pip install nibabel', or else install vreg with the "
    "rw option as 'pip install vreg[rw]'."
)
ZARR_IMPORT_ERROR = (
    "Saving in zarray format requires the zarr python package. You can "
    "install it with 'pip install zarr', or else install vreg with the "
    "rw option as 'pip install vreg[rw]'."
)


def write_npz(vol:Volume3D, filepath:str):
    """Write a volume to a single file in numpy's uncompressed .npz file format

    Args:
        vol (Volume3D): the volume to write.
        filepath (str): filepath to the .npz file.
    """
    np.savez(filepath, values=vol.values, affine=vol.affine)


def read_npz(filepath:str):
    """Load a volume created by write_npz()

    Args:
        filepath (str): filepath to the .npz file.

    Returns:
        Volume3D: the volume read from file.
    """
    npz = np.load(filepath)
    if 'values' not in npz.files:
        raise ValueError("The .npz file has not been created by write_npz.")
    if 'affine' not in npz.files:
        raise ValueError("The .npz file has not been created by write_npz.")
    return Volume3D(npz['values'], npz['affine'])


def _affine_to_from_RAH(affine):
    # convert to/from nifti coordinate system
    rot_180 = np.identity(4, dtype=np.float32)
    rot_180[:2,:2] = [[-1,0],[0,-1]]
    return np.matmul(rot_180, affine)


# TODO: Build a LazyVolume3D()??
# vol = vreg.from_zarr(path)
# vol.values is a zarray
# read Volume3D with vol01 = vol[i0:i1]
# write Colume3D with vol[i0:i1] = vol01


def create_zarr(vol:Volume3D, **kwargs):
    """Write volume to disk in zarray format.

    Args:
        vol (Volume3D): the volume to write.
        kwargs (dict): Any keyword arguments accepted by zarr.create()

    Raises:
        ImportError: Error raised if zarr is not installed.

    Note:
        This requires a separate installation of the optional zarr package, 
        either via 'pip install zarr' or by installing vreg with the rw 
        option 'pip install vreg[rw]'.
    """
    if not zarr_installed:
        raise ImportError(ZARR_IMPORT_ERROR)
    zarray = zarr.create(**kwargs)
    zarray[:] = vol.values
    zarray.attrs['affine'] = vol.affine.tolist()
    if vol.coords is not None:
        zarray.attrs['coords'] = [c.tolist() for c in vol.coords]
        zarray.attrs['dims'] = vol.dims
    return zarray


def write_nifti(vol:Volume3D, filepath:str):
    """Write volume to disk in NIfTI format.

    Args:
        vol (Volume3D): the volume to write.
        filepath (str): filepath to the NIfTI file.

    Raises:
        ImportError: Error raised if nibabel is not installed.

    Note:
        This requires a separate installation of the optional nibabel package, 
        either via 'pip install nibabel' or by installing vreg with the rw 
        option 'pip install vreg[rw]'.

    Warning:
        For volumes with more than 3 dimensions, saving to nifti will 
        result in loss of information on the extra dimensions - 
        specifically their name and coordinates of the images. 
    """
    if not nib_installed:
        raise ImportError(NIFTI_IMPORT_ERROR)
    affine = _affine_to_from_RAH(vol.affine)
    nifti = nib.Nifti1Image(vol.values, affine)
    nib.save(nifti, filepath)


def read_nifti(filepath:str, coords:np.ndarray=None, 
               dims:list=None, prec:int=None):
    """Read volume from a NIfTI file on disk.

    Args:
        filepath (str): filepath to the NIfTI file.
        coords (np.ndarray): For values with dimensions more than 3, 
            provide an additional coordinate array for the locations 
            of the replicated volumes. If not provided, index arrays 
            are used for the coordinates of the extra dimensions.
        dims (list): Names of the extra dimensions for volumes with more 
            than 3 dimensions.
        prec (int): internal precision to be used when comparing positions. 
            If not provided, the exact floating point values of locations 
            are used. Defaults to None.

    Raises:
        ImportError: Error raised if nibabel is not installed.

    Returns:
        Volume3D: the volume read from file.

    Note:
        This requires a separate installation of the optional nibabel package, 
        either via 'pip install nibabel' or by installing vreg with the rw 
        option 'pip install vreg[rw]'.

    Warning:
        For volumes with more than 3 dimensions, coordinates and 
        dimensions need to be provided separately as nifti does not 
        store those.
    """
    if not nib_installed:
        raise ImportError(NIFTI_IMPORT_ERROR)
    img = nib.load(filepath)
    values = img.get_fdata()
    affine = _affine_to_from_RAH(img.affine)
    return Volume3D(values, affine, coords, dims, prec)



import os
import sys
import pickle
import requests

from vreg import vol, rw

# filepaths need to be identified with importlib_resources
# rather than __file__ as the latter does not work at runtime
# when the package is installed via pip install

if sys.version_info < (3, 9):
    # importlib.resources either doesn't exist or lacks the files()
    # function, so use the PyPI version:
    import importlib_resources
else:
    # importlib.resources has files(), so use that:
    import importlib.resources as importlib_resources



DATASETS = [
    'Dixon_fat',
    'Dixon_water',
    'Dixon_in_phase',
    'Dixon_out_phase',
    'kidneys',
    'left_kidney',
    'right_kidney',
    'MTR',
    'T1_00',
    'T1_01',
    'T1_02',
    'T1_03',
    'T1_04',
    'T2star_00',
    'T2star_01',
    'T2star_02',
    'T2star_03',
    'T2star_04',
]


def fetch(dataset=None, clear_cache=False, download_all=False) -> vol.Volume3D:
    """Fetch a dataset included in vreg.

    The source data are on `Zenodo <https://doi.org/10.5281/zenodo.14630318>`_

    Args:
        dataset (str, optional): name of the dataset. See below for options.
        clear_cache (bool, optional): When a dataset is fetched, it is 
          downloaded and then stored in a local cache memory for faster access 
          next time it is fetched. Set clear_cache=True to delete all data 
          in the cache memory. Default is False.
        download_all (bool, optional): By default only the dataset that is 
          fetched is downloaded. Set download_all=True to download all 
          datasets at once. This will cost some time but then offers fast and 
          offline access to all datasets afterwards. This will take up around 
          300 MB of space on your hard drive. Default is False.

    Returns:
        Volume3D or list: Data as a vreg.Volume3D (for 3D scans) or a list of 
        Volume3D objects (for 2D multislice scans).

    Example:
        Fetch the kidneys mask and the T1-map, and display as overlay:

    .. pyvista-plot::
        :include-source:
        :context: 
        :force_static:

        >>> import vreg
        >>> import vreg.plot as plt

        Get the data:

        >>> kidneys = vreg.fetch('kidneys')
        >>> T1 = vreg.fetch('T1')

        Plot as overlays:

        >>> plt.overlay_2d(T1, kidneys)

    """

    if dataset is None:
        v = None
    elif dataset in ['T1', 'T2star']:
        v = [_fetch_dataset(f'{dataset}_{str(i).zfill(2)}') for i in range(5)]
    else:
        v = _fetch_dataset(dataset)

    if clear_cache:
        _clear_cache()

    if download_all:
        for d in DATASETS:
            _download(d)

    return v



def _clear_cache():
    """
    Clear the folder where the data downloaded via fetch are saved.

    Note if you clear the cache the data will need to be downloaded again 
    if you need them.
    """
    # TODO: This also deletes __init__.py
    f = importlib_resources.files('vreg.datafiles')
    for item in f.iterdir(): 
        if item.is_file(): 
            if str(item)[-11:] != '__init__.py':
                item.unlink() # Delete the file


def _fetch_dataset(dataset):

    f = importlib_resources.files('vreg.datafiles')
    datafile = str(f.joinpath(dataset + '.nii'))

    # If this is the first time the data are accessed, download them.
    if not os.path.exists(datafile):
        _download(dataset)

    return rw.read_nifti(datafile)


def _download(dataset):
        
    f = importlib_resources.files('vreg.datafiles')
    datafile = str(f.joinpath(dataset + '.nii'))

    if os.path.exists(datafile):
        return

    # Dataset location
    version_doi = "15535700" # This will change if a new version is created on zenodo
    file_url = "https://zenodo.org/records/" + version_doi + "/files/" + dataset + ".nii"

    # Make the request and check for connection error
    try:
        file_response = requests.get(file_url) 
    except requests.exceptions.ConnectionError as err:
        raise requests.exceptions.ConnectionError(
            "\n\n"
            "A connection error occurred trying to download the test data \n"
            "from Zenodo. This usually happens if you are offline. The \n"
            "first time a dataset is fetched via vreg.fetch you need to \n"
            "be online so the data can be downloaded. After the first \n"
            "time they are saved locally so afterwards you can fetch \n"
            "them even if you are offline. \n\n"
            "The detailed error message is here: " + str(err)) 
    
    # Check for other errors
    file_response.raise_for_status()

    # Save the file locally 
    with open(datafile, 'wb') as f:
        f.write(file_response.content)


import numpy as np
import vreg



def test_3d():

    shape = (4,5,2)
    vals = np.zeros(shape)
    affine = np.eye(4)
    vol = vreg.Volume3D(vals, affine)
    assert vol.shape == shape


def test_nd():

    shape = (4,5,2,3)
    vals = np.zeros(shape)
    affine = np.eye(4)
    vol = vreg.Volume3D(vals, affine)
    assert vol.shape == shape


def test_coords():

    shape = (4,5,2,3,9)
    vals = np.zeros(shape)
    affine = np.eye(4)

    # Default coordinates
    vol = vreg.Volume3D(vals, affine)
    assert vol.coords[0][-1,-1] == 2
    assert vol.coords[1][-1,-1] == 8

    # Well-formatted coordinates - same type
    coords = [np.zeros((3,9)), np.ones((3,9))]
    vol = vreg.Volume3D(vals, affine, coords)
    assert vol.coords[0][-1,-1] == 0
    assert vol.coords[1][-1,-1] == 1

    # Well-formatted coordinates - mixed type
    coords = [np.full((3,9), 'A'), np.ones((3,9))]
    vol = vreg.Volume3D(vals, affine, coords)
    assert vol.coords[0][-1,-1] == 'A'
    assert vol.coords[1][-1,-1] == 1

    # Ill-formatted coordinates (single coordinate, needs two)
    coords = [np.zeros((3,9))]
    try:
        vol = vreg.Volume3D(vals, affine, coords)
    except:
        assert True
    else:
        assert False

    # Ill-formatted coordinates (scalar)
    coords = np.zeros((3,9))
    try:
        vol = vreg.Volume3D(vals, affine, coords)
    except:
        assert True
    else:
        assert False

    # Ill-formatted coordinates (one scalar)
    coords = ['A', np.ones((3,9))]
    try:
        vol = vreg.Volume3D(vals, affine, coords)
    except:
        assert True
    else:
        assert False

    # Ill-formatted coordinates (wrong dimensions)
    coords = [np.full((2,9), 'A'), np.ones((3,9))]
    try:
        vol = vreg.Volume3D(vals, affine, coords)
    except:
        assert True
    else:
        assert False


def test_copy():
    shape = (4,5,2,3,9)
    vals = np.zeros(shape)
    affine = np.eye(4)  
    coords = [np.full((3,9), 'A'), np.ones((3,9))]
    dims = ['Name', 'Pos']
    vol = vreg.Volume3D(vals, affine, coords, dims)  
    vol_copy = vol.copy()
    vol_copy.coords[0] = np.full((3,9), 'B')
    assert vol.coords[0][-1,-1] == 'A'
    assert vol_copy.coords[0][-1,-1] == 'B'


def test_separate():

    # Build test volume
    shape = (4,5,2,3,9)
    vals = np.zeros(shape)
    affine = np.eye(4)  
    coords = [np.full((3,9), 'A'), np.ones((3,9))]
    coords[0][1,:] = 'B'
    coords[0][2,:] = 'C'
    dims = ['Name', 'Pos']
    vol = vreg.Volume3D(vals, affine, coords, dims) 

    # Separate along 1 axis
    vols_sep = vol.separate(axis=3)    
    assert vols_sep.shape == (3,)
    assert vols_sep[0].shape == (4,5,2,1,9)
    assert vols_sep[0].coords[0][0,8] == 'A'
    assert vols_sep[1].coords[0][0,8] == 'B'
    assert vols_sep[2].coords[0][0,8] == 'C'

    # Separate along all axes
    vols_sep = vol.separate()  
    assert vols_sep.shape == (3,9)
    assert vols_sep[0,8].shape == (4,5,2,1,1)
    assert vols_sep[0,8].coords[0][0,0] == 'A'
    assert vols_sep[1,8].coords[0][0,0] == 'B'
    assert vols_sep[2,8].coords[0][0,0] == 'C'


if __name__ == '__main__':

    test_3d()
    test_nd()
    test_coords()
    test_copy()
    test_separate()

    print('All init tests passed!!!')


import numpy as np
import scipy.ndimage as ndi

import vreg.utils



# def mask_volume(array, affine, array_mask, affine_mask, margin:float=0):

#     # Overlay the mask on the array.
#     array_mask, _ = affine_reslice(array_mask, affine_mask, affine, array.shape)

#     # Mask out array pixels outside of region.
#     array *= array_mask

#     # Extract bounding box around non-zero pixels in masked array.
#     array, affine = vreg.utils.bounding_box(array, affine, margin)
    
#     return array, affine

def mask_volume(array, affine, array_mask, affine_mask, margin:float=0):

    # Overlay the mask on the array.
    array_mask, _ = affine_reslice(array_mask, affine_mask, affine, array.shape)

    # Extract bounding box around non-zero pixels in masked array.
    geom = vreg.utils.bounding_box_geom(array_mask, affine, margin)
    array, affine = vreg.utils.bounding_box_apply(array, affine, *geom)
    
    return array, affine



def apply_affine(affine, coord):
    """Apply affine transformation to an array of coordinates"""

    nd = affine.shape[0]-1
    matrix = affine[:nd,:nd]
    offset = affine[:nd, nd]
    return np.dot(coord, matrix.T) + offset
    #return np.dot(matrix, co.T).T + offset


def apply_inverse_affine(
        input_data, inverse_affine, 
        output_shape, output_coordinates=None, 
        order=1, **kwargs):

    # Create an array of all coordinates in the output volume
    if output_coordinates is None:
        output_coordinates = vreg.utils.volume_coordinates(output_shape)

    # Apply affine transformation to all coordinates in the output volume
    # nd = inverse_affine.shape[0]-1
    # matrix = inverse_affine[:nd,:nd]
    # offset = inverse_affine[:nd, nd]
    # input_coordinates = np.dot(output_coordinates, matrix.T) + offset
    # #co = np.dot(matrix, co.T).T + offset
    input_coordinates = apply_affine(inverse_affine, output_coordinates)

    # Extend with constant value for half a voxel outside of the boundary
    input_coordinates = vreg.utils.extend_border(
        input_coordinates, input_data.shape)

    # Interpolate the volume in the transformed coordinates
    #output_data = ndi.map_coordinates(input_data, input_coordinates.T, **kwargs)
    output_data = ndi.map_coordinates(
        input_data, input_coordinates.T, order=order, **kwargs)
    output_data = np.reshape(output_data, output_shape)

    return output_data


def affine_reslice(
        input_data, input_affine:np.ndarray, 
        output_affine:np.ndarray, output_shape=None, 
        **kwargs):

    # If the output shape is 2d, add a 3d dimension of size 1 for the calculation
    output_2d = False
    if output_shape is not None:
        output_2d = np.size(output_shape)==2
        if output_2d:
            output_shape = output_shape + (1,)

    # If 2d array, add a 3d dimension of size 1
    if input_data.ndim == 2: 
        input_data = np.expand_dims(input_data, axis=-1)

    # If no output shape is provided, retain the physical volume of the input datas
    if output_shape is None:

        # Get field of view from input data
        input_pixel_spacing = np.linalg.norm(input_affine[:3,:3], axis=0)
        field_of_view = np.multiply(np.array(input_data.shape), input_pixel_spacing)

        # Find output shape for the same field of view
        #output_rotation, output_translation, output_pixel_spacing = vreg.utils.affine_components(output_affine)
        output_pixel_spacing = np.linalg.norm(output_affine[:3,:3], axis=0)
        output_shape = np.around(np.divide(field_of_view, output_pixel_spacing)).astype(np.int16)
        output_shape[np.where(output_shape==0)] = 1

        # Adjust output pixel spacing to fit the field of view
        output_pixel_spacing = np.divide(field_of_view, output_shape)
        new_output_affine = output_affine.copy()
        for d in [0,1,2]:
            new_output_affine[:3,d] = output_pixel_spacing[d] * output_affine[:3,d]/np.linalg.norm(output_affine[:3,d])
        output_affine = new_output_affine
        # output_affine = vreg.utils.affine_matrix(
        #     rotation=output_rotation, translation=output_translation, 
        #     pixel_spacing=output_pixel_spacing)

    # Reslice input data to output geometry
    transform = np.linalg.inv(input_affine).dot(output_affine) # Ai B
    output_data = apply_inverse_affine(input_data, transform, output_shape, **kwargs)

    # If a 2d output shape was requested, return a 2d array
    if output_2d:
        output_data = output_data[:,:,0]

    return output_data, output_affine


def affine_transform(input_data, input_affine, transformation, reshape=False, **kwargs):

    # If 2d array, add a 3d dimension of size 1
    if input_data.ndim == 2: 
        input_data = np.expand_dims(input_data, axis=-1)

    if reshape:
        output_shape, output_affine = vreg.utils.affine_output_geometry(input_data.shape, input_affine, transformation)
    else:
        output_shape, output_affine = input_data.shape, input_affine.copy()

    # Perform the inverse transformation
    affine_transformed = transformation.dot(input_affine)
    inverse = np.linalg.inv(affine_transformed).dot(output_affine) # Ainv Tinv B 
    output_data = apply_inverse_affine(input_data, inverse, output_shape, **kwargs)

    return output_data, output_affine

# This needs a
def affine_transform_and_reslice(input_data, input_affine, output_shape, output_affine, transformation, **kwargs):

    input_data = vreg.utils.to_3d(input_data)
    affine_transformed = transformation.dot(input_affine)
    inverse = np.linalg.inv(affine_transformed).dot(output_affine) # Ai Ti B
    output_data = apply_inverse_affine(input_data, inverse, output_shape, **kwargs)

    return output_data



import numpy as np

# default metrics
# ---------------


def cov_mask(static, transformed, nan=None):
    # transformed is here a mask image.
    if nan is None:
        masked = static[transformed > 0.01]
    else:
        masked = static[(transformed > 0.01) & (transformed != nan)]
    return np.std(masked)/np.mean(masked)


def sum_of_squares(static, transformed, nan=None):
    if nan is not None:
        i = np.where(transformed != nan)
        st, tr = static[i], transformed[i]
    else:
        st, tr = static, transformed
    return np.sum(np.square(st-tr))
    

def mutual_information(static, transformed, nan=None):

    # Mask if needed
    if nan is not None:
        i = np.where(transformed != nan)
        st, tr = static[i], transformed[i]
    else:
        st, tr = static, transformed
    # Calculate 2d histogram
    hist_2d, _, _ = np.histogram2d(st.ravel(), tr.ravel(), bins=20)
    # Convert bins counts to probability values
    pxy = hist_2d / float(np.sum(hist_2d))
    px = np.sum(pxy, axis=1) # marginal for x over y
    py = np.sum(pxy, axis=0) # marginal for y over x
    px_py = px[:, None] * py[None, :] # Broadcast to multiply marginals
    # Now we can do the calculation using the pxy, px_py 2D arrays
    nzs = pxy > 0 # Only non-zero pxy values contribute to the sum
    return -np.sum(pxy[nzs] * np.log(pxy[nzs] / px_py[nzs]))


def mi_grad(static, transformed, nan=None):
    gstatic = np.gradient(np.squeeze(static))
    gtransf = np.gradient(np.squeeze(transformed))
    gstatic = np.linalg.norm(np.stack(gstatic), axis=0)
    gtransf = np.linalg.norm(np.stack(gtransf), axis=0)
    # support = np.squeeze(static) > 1e-2
    # gstatic = gstatic[support]
    # gtransf = gtransf[support]
    # gstatic /= np.linalg.norm(gstatic)
    # gtransf /= np.linalg.norm(gtransf)
    return mutual_information(gstatic, gtransf, nan=nan)

def sos_grad(static, transformed, nan=None):
    gstatic = np.gradient(np.squeeze(static))
    gtransf = np.gradient(np.squeeze(transformed))
    gstatic = np.linalg.norm(np.stack(gstatic), axis=0)
    gtransf = np.linalg.norm(np.stack(gtransf), axis=0)
    gstatic /= np.linalg.norm(gstatic)
    gtransf /= np.linalg.norm(gtransf)
    return sum_of_squares(gstatic, gtransf, nan=nan)


import numpy as np

from vreg import mod_affine, metrics, utils, optimize


def _check_dims(dims, shape):
    if len(dims) != len(shape)-3:
        raise ValueError(
            f"dims should have {len(shape)-3} values."
        )  

def _check_coords(coords, shape):

    if not isinstance(coords, list):
        raise ValueError("coords must be a list of arrays")
    if len(coords) != len(shape)-3:
        raise ValueError(
            f"coords should have {len(shape)-3} elements."
        )
    for coord in coords:
        if not isinstance(coord, np.ndarray):
            raise ValueError(
                'All coordinates should be numpy arrays.'
            )
        if coord.shape != shape[3:]:
            raise ValueError(
                f"All coordinates should have dimensions {shape[3:]}."
            )

class Volume3D:
    # TODO: for multidimensional volumes, one affine per volume
    # API ND:
    # values[:, :, :, i, j]
    # affine[:, :, i, j] -> Allow each volume to have its own affine?
    # coords[:][i, j]
    # dims[:]
    # API 3D:
    # values[:, :, :]
    # affine[:, :]
    # coords = None
    # dims = None
    """
    A spatially aware numpy array
    
    Args:
        values (np.ndarray): Numpy array of values (3 or more dimensions)
        affine (np.ndarray): 4x4 numpy array with the affine matrix 
            of the value array. If not provided, the identity is assumed. 
            Defaults to None.  
        coords (list): For values with more than 3 dimensions, 
            provide a list of coordinate arrays for the locations 
            of the replicated volumes. If not provided, index arrays 
            are used for the coordinates of the extra dimensions.
        dims (list): Names of the extra dimensions for volumes with more 
            than 3 dimensions.
        prec (int): internal precision to be used when comparing positions. 
            If not provided, the exact floating point values of locations 
            are used. Defaults to None.
    """

    def __init__(self, values:np.ndarray, affine:np.ndarray, 
                 coords:list=None, dims:list=None, prec:int=None):
        
        # Set precision
        if prec is not None:
            if not isinstance(prec, int):
                raise ValueError("prec needs to be an integer.")
        self._prec = prec

        # Set values
        if not isinstance(values, np.ndarray):
            raise ValueError('values must be a numpy array.')
        if values.ndim < 3:
            raise ValueError("values must have at least 3 dimensions.")
        self._values = values

        # Set affine
        if not isinstance(affine, np.ndarray):
            raise ValueError('affine must be a numpy array.')
        if affine.shape != (4,4):
            raise ValueError('affine must be a 4x4 array.')
        self._affine = affine

        # Set coords & dims
        self._coords = None
        self._dims = None
        if values.ndim > 3:

            # Set dims
            if dims is None:
                dims = list(range(values.ndim-3))
            _check_dims(dims, values.shape)  
            self._dims = dims

            # Set coords
            if coords is None:
                coords = [np.arange(d) for d in values.shape[3:]]
                coords = np.meshgrid(*coords, indexing='ij')
                coords = list(coords)
            _check_coords(coords, values.shape)
            self._coords = coords
            
    
    @property
    def values(self):
        return self._values
    
    @property
    def affine(self):
        return self._affine
    
    @property
    def coords(self):
        return self._coords
    
    @property
    def dims(self):
        return self._dims
    
    @property
    def prec(self):
        return self._prec
    
    @property
    def ndim(self): 
        return self._values.ndim
    
    @property
    def shape(self):
        return self._values.shape
    
    @property
    def spacing(self):
        return np.linalg.norm(self.affine[:3,:3], axis=0)
    
    @property
    def row_dir(self):
        return self.affine[:3,0]/np.linalg.norm(self.affine[:3,0])
    
    @property
    def col_dir(self):
        return self.affine[:3,1]/np.linalg.norm(self.affine[:3,1])
    
    @property
    def slice_dir(self):
        return self.affine[:3,2]/np.linalg.norm(self.affine[:3,2])
    
    @property
    def pos(self):
        return self.affine[:3,3]
    
    @property
    def is_right_handed(self):
        normal = np.cross(self.affine[:3,0], self.affine[:3,1])
        proj = np.dot(self.affine[:3,2], normal)
        return proj > 0


    def loc(self, axis) -> float:
        """Location of the volume along an axis

        Args:
            axis (int): either 0 (x-axis), 1 (y-axis) or 2 (z-axis)

        Returns:
            float: location along the specified axis
        """
        # project the position onto the axis direction
        axis_dir = self.affine[:3, axis]/np.linalg.norm(self.affine[:3, axis])
        return np.dot(self.affine[:3, 3], axis_dir)


    def set_values(self, values:np.ndarray):
        """Set new values for the volume

        Args:
            values (np.ndarray): array with the same shape as the 
                existing array of values.
        """
        if not isinstance(values, np.ndarray):
            raise ValueError('values must be a numpy array.')
        
        if values.shape != self.values.shape:
            raise ValueError(f'values must have dimensions {self.values.shape}.')
        
        self._values = values
        return self


    def set_affine(self, affine:np.ndarray):
        """Set a new affine for the volume

        Args:
            affine (np.ndarray): 4x4 affine array
        """
        if not isinstance(affine, np.ndarray):
            raise ValueError('affine must be a numpy array.')
        
        if affine.shape != (4,4):
            raise ValueError('affine must be a 4x4 array.')
        
        self._affine = affine
        return self


    def set_coords(self, coords:np.ndarray):
        """Set new coordinates for the volume

        Args:
            coords (np.ndarray): Array of coordinates

        Raises:
            ValueError: If the new coords have the wrong shape.
        """
        _check_coords(coords, self.shape)
        self._coords = coords
        return self


    def set_dims(self, dims:list):
        """Set new names for the non-spatial dimensions

        Args:
            dims (np.ndarray): List of dimension names

        Raises:
            ValueError: If the new dims has the wrong length.
        """
        _check_dims(dims, self.shape)    
        self._dims = dims
        return self

    
    def copy(self, **kwargs):
        """Return a copy

        Args:
            kwargs: Any keyword arguments accepted by `numpy.copy`.

        Returns:
            Volume3D: copy
        """
        return Volume3D(
            self.values.copy(**kwargs), 
            self.affine.copy(**kwargs),
            [c.copy(**kwargs) for c in self.coords],
            list(np.asarray(self.dims).copy(**kwargs)),
            self._prec,
        )
    
    def extract_slice(self, z=0):
        """Extract a slice at index z

        Args:
            z (int, optional): slice index

        Returns:
            vreg.Volume3D: one-slice Volume3D at index z
        """
        if self.ndim > 3:
            raise ValueError("This function is not yet available for volumes "
                             "with more than 3 dimensions")
        values, affine = utils.extract_slice(self.values, self.affine, z)
        values = np.expand_dims(values.copy(), axis=-1)
        return Volume3D(values, affine.copy())


    def separate(self, axis=None):
        """Separate a volume along one or more time axes.

        Args:
            axis (int, optional): Time axis along which to separate. If this 
              is not provided, the volume is separated along all time 
              axes. Defaults to None.

        Returns:
            np.ndarray: array of 3D volumes. 
        """

        if axis in [0,1,2]:
            raise ValueError("A volume cannot be separated along a space axis. "
                             "Use split() instead.")  
        if self.ndim==3:
            raise ValueError("A 3D volume can not be separated. Perhaps "
                              "you wanted to split()?")
        if axis is not None:

            # Separate along a prespecified dimensions
            vols = []
            for i in range(self.shape[axis]):
                coords_i = [_take_view_keepdims(c, i, axis-3) for c in self.coords]
                affine_i = self.affine
                values_i = _take_view_keepdims(self.values, i, axis)
                # Build the volume and add to the list.
                vol_i = Volume3D(values_i, affine_i, coords_i, self.dims) 
                vols.append(vol_i)
            return np.asarray(vols) 
        
        else:
            # Separate along all dimensions
            axis = 3
            vols = self.separate(axis) # 1D
            if axis < self.ndim-1:
                axis += 1
                shape = vols.shape
                vols = [v.separate(axis) for v in vols.reshape(-1)]
                newdim = vols[0].size
                vols = np.array(vols).reshape(shape + (newdim,)) # 3D
            return vols


    def split(self, n=None, axis=2, gap=0)->list:
        """Split a volume into slices (2D volumes)

        Args:
            n (int, optional): number of slices in the result. If this is not 
              provided, n is the shape of the volume in the axis along which 
              it is split. Defaults to None.
            axis (int, optional): Axis along which to split the volume. 
              Defaults to -1 (z-axis).
            gap (float, optional): Add a gap (in mm) between the resulting 
              slices. Defaults to 0.

        Returns:
            list of Volume3D: a list of volumes with a single slice each.
        """
        if axis not in [0,1,2]:
            raise ValueError("A volume cannot be split along a time axis. Use "
                             "separate() instead.")           
        
        # Default
        if n is None:
            n = self.shape[axis]
        if n==1:
            return [self]
        # If the number of slices required is different from the current 
        # number of slices, then first resample the volume.
        if n == self.shape[axis]:
            vol = self
        else:
            spacing = self.spacing
            spacing[axis] = spacing[axis]*self.shape[axis]/n
            vol = self.resample(spacing) 

        # Split up the volume into a list of 2D volumes.
        mat = vol.affine[:3,:3]
        split_vec = mat[:, axis] 
        split_unit_vec = split_vec/self.spacing[axis]
        vols = []
        for i in range(vol.shape[axis]):
            
            # Shift the affine by i positions along the slice axis.
            affine_i = vol.affine.copy()
            affine_i[:3, 3] += i*split_vec + i*gap*split_unit_vec

            # Take the i-th slice and keep dimensions
            values_i = _take_view_keepdims(vol.values, i, axis)

            # Build the volume and add to the list.
            vol_i = Volume3D(values_i, affine_i, self.coords, self.dims)
            vols.append(vol_i)
        return vols

    
    def add(self, v, *args, **kwargs):
        """Add another volume

        Args:
            v (Volume3D): volume to add. If this is in a different geometry, it 
                will be resliced first
            args, kwargs: arguments and keyword arguments of `numpy.add`.

        Returns:
            Volume3D: sum of the two volumes
        """
        if self.ndim > 3:
            raise ValueError("This function is not yet available for volumes "
                             "with more than 3 dimensions")
        v = v.slice_like(self)
        values = np.add(self.values, v.values, *args, **kwargs)
        return Volume3D(values, self.affine)
    
    def subtract(self, v, *args, **kwargs):
        """Subtract another volume

        Args:
            v (Volume3D): volume to subtract. If this is in a different 
                geometry, it will be resliced first
            args, kwargs: arguments and keyword arguments of `numpy.subtract`.

        Returns:
            Volume3D: sum of the two volumes
        """
        if self.ndim > 3:
            raise ValueError("This function is not yet available for volumes "
                             "with more than 3 dimensions")
        v = v.slice_like(self)
        values = np.subtract(self.values, v.values, *args, **kwargs)
        return Volume3D(values, self.affine)

    def crop(self, mask=None, margin=0.0): # better term than bounding_box
        """Crop to a box

        Args:
            mask (Volume3D, optional): If mask is None, the volume 
                is cropped to the non-zero values of the Volume3D. If mask is 
                provided, it is cropped to the non-zero values of mask 
                instead. Defaults to None.
            margin (float, optional): How big a margin (in physical units) 
                around the object. Defaults to 0.

        Returns:
            Volume3D: the bounding box
        """
        return self.bounding_box(mask, margin)
    

    def bounding_box(self, mask=None, margin=0.0): # replace by crop
        """Return the bounding box

        Args:
            mask (Volume3D, optional): If mask is None, the bounding box is 
                drawn around the non-zero values of the Volume3D. If mask is 
                provided, it is drawn around the non-zero values of mask 
                instead. Defaults to None.
            margin (float, optional): How big a margin (in mm) around the 
                object. Defaults to 0.

        Returns:
            Volume3D: the bounding box
        """
        if self.ndim > 3:
            raise ValueError("This function is not yet available for volumes "
                             "with more than 3 dimensions")
        if mask is None:
            values, affine = mod_affine.mask_volume(
                self.values, self.affine, self.values, self.affine, margin)
        else:
            values, affine = mod_affine.mask_volume(
                self.values, self.affine, mask.values, mask.affine, margin)
        return Volume3D(values, affine)


    def resample(self, spacing=None, stretch=None):
        """Resample volume to new pixel spacing

        Args:
            spacing (array-like, optional): New pixel spacing in mm. Generally 
                this is a 3-element array but for isotropic resampling this can 
                be a scalar value. If this is not provided, the volume is 
                resampled according to the specified stretch. Defaults to None.
            stretch (float, optional): Rescale pixel size with this value. 
                Generally this is a 3-element array, one for each dimension. If 
                a scalar value is provided, all dimensions are resampled with 
                the same stretch factor. This argument is ignored if a spacing is 
                provided explicitly. Defaults to None (no resampling).

        Raises:
            ValueError: if spacing or stretch have the wrong size.

        Returns:
            vreg.Volume3D: resampled volume
        """
        # Set defaults
        if stretch is None:
            stretch = 1.0

        # Affine components
        #rot, trans, ps = utils.affine_components(self.affine)
        ps = self.spacing

        # Get new pixel spacing
        if spacing is None:
            if np.isscalar(stretch):
                spacing = ps*stretch
            elif np.size(stretch)==3:
                spacing = ps*stretch
            else:
                raise ValueError(
                    'stretch must be a scalar or a 3-element array')
        elif np.isscalar(spacing):
            spacing = np.full(3, spacing)
        elif np.size(spacing) != 3:
            raise ValueError(
                'spacing must be a scalar or a 3-element array')
        
        # Resample
        affine = self.affine.copy()
        for d in [0,1,2]:
            affine[:3, d] = spacing[d]*self.affine[:3,d]/self.spacing[d]
        # affine = utils.affine_matrix(rotation=rot, translation=trans, 
        #                              pixel_spacing=spacing)
        if len(self.shape)==3:
            values_reslice, _ = mod_affine.affine_reslice(self.values, self.affine, affine)
        else:
            values_reshape = self.values.reshape(self.shape[:3] + (-1,))
            values_reslice_0, _ = mod_affine.affine_reslice(values_reshape[...,0], self.affine, affine)
            values_reslice = np.zeros(values_reslice_0.shape + (values_reshape.shape[-1],))
            values_reslice[...,0] = values_reslice_0
            for d in range(1, values_reshape.shape[-1]):
                values_reslice[...,d], _ = mod_affine.affine_reslice(values_reshape[...,d], self.affine, affine)
            values_reslice = values_reslice.reshape(values_reslice_0.shape + self.shape[3:])
       
        return Volume3D(values_reslice, affine, self.coords, self.dims)
    

    def find_transform_to(
            self, target, transform, params=None, metric=None, 
            optimizer=None, resolutions=None, 
            mask=None, target_mask=None, **kwargs):
        """Coregister a volume to a static target volume.

        Args:
            target (vreg.Volume3D): Fixed target volume for the coregistration.
            transform (str, optional): Coordinate transformation. 
              Possible values are 'translate', 'rotate', 'stretch', 
              'transform_rigid', 'transform_affine'.
            params (array-like, optional): Initial parameters of the 
              transformation. Defaults to None.
            metric (func, optional): Metric to quantify goodness of alignment. 
              Options are 'mi' (mutual information), 'sos' (sum of squares), 
              and 'migrad' (mutual information of the image gradient).  
              Default is 'mi'.
            optimizer (dict, optional): Optimizer as a dictionary 
              with one key *method* specifying the method used for optimization. 
              The other items in the dictionary are any optional keywords 
              accepted by the method. Defaults to {'method':'LS'}.
            resolutions (list, optional): Resolutions to use in the 
              optimization. Defaults to [1].
            mask (Volume3D, optional): volume used for masking the moving 
              volume. Defaults to None.
            target_mask (Volume3D, optional): volume used for masking the 
              static volume. Defaults to None.
            kwargs (dict, optional): optional keyword arguments for the 
              transform function.

        Returns:
            params: The optimal values for the transformaton parameters.
        """
        if self.ndim > 3:
            raise ValueError("This function is not yet available for volumes "
                             "with more than 3 dimensions")
        # Defaults
        if metric is None:
            metric = 'mi'
        if optimizer is None:
            optimizer = {'method':'LS'}
        if resolutions is None:
            resolutions = [1]

        # Perform multi-resolution loop
        mask_resampled = None
        target_ind = None

        for res in resolutions:

            if res == 1:
                moving = self
                target_resampled = target
            else:
                # Downsample
                moving = self.resample(stretch=res)
                target_resampled = target.resample(stretch=res)

            # resample the masks
            if mask is not None:
                mask_resampled = mask.slice_like(moving)
            if target_mask is not None:
                target_mask_resampled = target_mask.slice_like(target_resampled)
                target_ind = np.where(target_mask_resampled.values >= 0.5)

            args = (
                target_resampled, transform, metric, mask_resampled, 
                target_ind, kwargs,
            )
            params = optimize.minimize(
                moving._dist, params, args=args, **optimizer)
            
        return params

    def _dist(self, params, target, transform, metric, mask, target_ind, 
              kwargs):
        return self.distance(target, transform, params, metric, mask, 
                                   target_ind, **kwargs)

    def distance(self, target, transform, params, metric='mi', mask=None, 
                       target_ind=None, **kwargs):
        """Distance to a target volume after a transform

        Args:
            target (vreg.Volume3D): Target volume
            transform (str, optional): Coordinate transformation. 
              Possible values are 'translate', 'rotate', 'stretch', 
              'transform_rigid', 'transform_affine'.
            params (array-like, optional): Initial parameters of the 
              transformation. Defaults to None.
            metric (func, optional): Metric to quantify distance. 
              Options are 'mi' (mutual information), 'sos' (sum of squares), 
              and 'migrad' (mutual information of the image gradient).  
              Default is 'mi'.
            mask (Volume3D, optional): volume used for masking the moving 
              volume. Defaults to None.
            target_ind (numpy.ndarray, optional): Indices in the target 
              volume that count towards the distance. Defaults to None.

        Returns:
            float: distance after transform
        """
        if self.ndim > 3:
            raise ValueError("This function is not yet available for volumes "
                             "with more than 3 dimensions")
        if metric == 'mi':
            metric = metrics.mutual_information
        elif metric == 'sos':
            metric = metrics.sum_of_squares
        elif metric == 'migrad':
            metric = metrics.mi_grad

        # Transform the moving image to the target
        moving = self.transform_to(target, transform, params, **kwargs)
        
        # Transform the moving mask
        mask_ind = None
        if mask is not None:
            mask = mask.transform_to(target, transform, params, **kwargs)
            mask_ind = np.where(mask >= 0.5)

        # Calculate metric in indices exposed by the mask(s)
        if target_ind is None and mask_ind is None:
            return metric(target.values, moving.values)
        
        if target_ind is None and mask_ind is not None:
            ind = mask_ind
        elif target_ind is not None and mask_ind is None:
            ind = target_ind
        elif target_ind is not None and mask_ind is not None:
            ind = target_ind or mask_ind
        return metric(target.values[ind], moving.values[ind])
    
        
    def transform_to(self, target, transform, params, **kwargs):
        """Transform a volume to a target volume

        Args:
            target (vreg.Volume3D): Target volume
            transform (str, optional): Coordinate transformation. 
              Possible values are 'translate', 'rotate', 'stretch', 
              'transform_rigid', 'transform_affine'.
            params (array-like, optional): Initial parameters of the 
              transformation. Defaults to None.

        Returns:
            vreg.Volume3D: transformed volume
        """
        if self.ndim > 3:
            raise ValueError("This function is not yet available for volumes "
                             "with more than 3 dimensions")
        return getattr(self, transform + '_to')(target, params, **kwargs)
    
    def transform(self, transform, params, **kwargs):
        """Transform a volume

        Args:
            transform (str, optional): Coordinate transformation. 
              Possible values are 'translate', 'rotate', 'stretch', 
              'transform_rigid', 'transform_affine'.
            params (array-like, optional): Initial parameters of the 
              transformation. Defaults to None.

        Returns:
            vreg.Volume3D: transformed volume
        """
        if self.ndim > 3:
            raise ValueError("This function is not yet available for volumes "
                             "with more than 3 dimensions")
        return getattr(self, transform)(params, **kwargs)


    def find_transform_rigid_to(
            self, target, params=None, metric=None, 
            optimizer=None, resolutions=None, 
            mask=None, target_mask=None, center=None, coords='fixed'):
        """Find the rigid transform to a static target volume.

        Args:
            target (vreg.Volume3D): Fixed target volume for the coregistration.
            params (array-like): 6-element vector with translation and  
              rotation vectors, in that order. 
            metric (func, optional): Metric to quantify goodness of alignment. 
              Options are 'mi' (mutual information), 'sos' (sum of squares), 
              and 'migrad' (mutual information of the image gradient).  
              Default is 'mi'.
            optimizer (dict, optional): Optimizer as a dictionary 
              with one key *method* specifying the method used for optimization. 
              The other items in the dictionary are any optional keywords 
              accepted by the method. Defaults to {'method':'LS'}.
            resolutions (list, optional): Resolutions to use in the 
              optimization. Defaults to [1].
            mask (Volume3D, optional): volume used for masking the moving 
              volume. Defaults to None.
            target_mask (Volume3D, optional): volume used for masking the 
              static volume. Defaults to None.
            center (str or array-like): center of rotation. If this has value 
              'com', the rotation is performed around the center of mass. 
              Alternatively this can be a 3-element vector. Defaults to None 
              (center = origin).
            coords (str or array, optional): Reference frame for the 
              coordinates of the translation, rotation and center vector, as 
              a 4x4 affine 
              array. String options are shorthand notations: 'fixed' (patient 
              reference frame), and 'volume' (volume reference frame). Defaults 
              to 'fixed'.

        Returns:
            params: The optimal values for the transformaton parameters.
        """
        if self.ndim > 3:
            raise ValueError("This function is not yet available for volumes "
                             "with more than 3 dimensions")
        return self.find_transform_to(
            target, 'transform_rigid', params=params, metric=metric, 
            optimizer=optimizer, resolutions=resolutions, 
            mask=mask, target_mask=target_mask, center=center, coords=coords)

    def find_rotate_to(
            self, target, rotation=None, metric=None, 
            optimizer=None, resolutions=None, 
            mask=None, target_mask=None, center=None, coords='fixed'):
        """Find the rigid transform to a static target volume.

        Args:
            target (vreg.Volume3D): Fixed target volume for the coregistration.
            rotation (array-like): 3-element rotation vector. 
            metric (func, optional): Metric to quantify goodness of alignment. 
              Options are 'mi' (mutual information), 'sos' (sum of squares), 
              and 'migrad' (mutual information of the image gradient).  
              Default is 'mi'.
            optimizer (dict, optional): Optimizer as a dictionary 
              with one key *method* specifying the method used for optimization. 
              The other items in the dictionary are any optional keywords 
              accepted by the method. Defaults to {'method':'LS'}.
            resolutions (list, optional): Resolutions to use in the 
              optimization. Defaults to [1].
            mask (Volume3D, optional): volume used for masking the moving 
              volume. Defaults to None.
            target_mask (Volume3D, optional): volume used for masking the 
              static volume. Defaults to None.
            center (str or array-like): center of rotation. If this has value 
              'com', the rotation is performed around the center of mass. 
              Alternatively this can be a 3-element vector. Defaults to None 
              (center = origin).
            coords (str or array, optional): Reference frame for the 
              coordinates of the translation, rotation and center vector, as 
              a 4x4 affine 
              array. String options are shorthand notations: 'fixed' (patient 
              reference frame), and 'volume' (volume reference frame). Defaults 
              to 'fixed'.

        Returns:
            params: The optimal values for the transformaton parameters.
        """
        if self.ndim > 3:
            raise ValueError("This function is not yet available for volumes "
                             "with more than 3 dimensions")
        return self.find_transform_to(
            target, 'rotate', params=rotation, metric=metric, 
            optimizer=optimizer, resolutions=resolutions, 
            mask=mask, target_mask=target_mask, center=center, coords=coords)
    
    
    def find_translate_to(
            self, target, translation=None, metric=None, 
            optimizer=None, resolutions=None, 
            mask=None, target_mask=None, coords='fixed'):
        """Find the translation to a static target volume.

        Args:
            target (vreg.Volume3D): Fixed target volume for the coregistration.
            translation (array-like, optional): Initial values for the 
              translation vector. Defaults to None.
            metric (func, optional): Metric to quantify goodness of alignment. 
              Options are 'mi' (mutual information), 'sos' (sum of squares), 
              and 'migrad' (mutual information of the image gradient).  
              Default is 'mi'.
            optimizer (dict, optional): Optimizer as a dictionary 
              with one key *method* specifying the method used for optimization. 
              The other items in the dictionary are any optional keywords 
              accepted by the method. Defaults to {'method':'LS'}.
            resolutions (list, optional): Resolutions to use in the 
              optimization. Defaults to [1].
            mask (Volume3D, optional): volume used for masking the moving 
              volume. Defaults to None.
            target_mask (Volume3D, optional): volume used for masking the 
              static volume. Defaults to None.
            coords (str or array, optional): Reference frame for the 
              coordinates of the translation vector, as a 4x4 affine 
              array. String options are shorthand notations: 'fixed' (patient 
              reference frame), and 'volume' (volume reference frame). Defaults 
              to 'fixed'.

        Returns:
            params: The optimal values for the transformaton parameters.
        """
        if self.ndim > 3:
            raise ValueError("This function is not yet available for volumes "
                             "with more than 3 dimensions")
        return self.find_transform_to(
            target, 'translate', params=translation, metric=metric, 
            optimizer=optimizer, resolutions=resolutions, 
            mask=mask, target_mask=target_mask, coords=coords)


    def _affine_matrix(self, translation, rotation, center, stretch, coords):
        # Helper function

        # Initialize
        if translation is None:
            translation = np.zeros(3)
        if rotation is None:
            rotation = np.zeros(3)
        if stretch is None:
            stretch = np.ones(3)
        if center is None:
            center = np.zeros(3)

        # Check inputs
        if 0 != np.count_nonzero(stretch <= 0):
            raise ValueError(
                "All elements of stretch must be strictly positive")
        if isinstance(coords, str):
            if coords not in ['fixed','volume']:
                raise ValueError(
                    coords + " is not a valid reference frame. The options "
                    "are 'volume' and 'fixed'.")
        elif np.shape(coords) != (4,4):
            raise ValueError(
                "coords must either be a string or an affine array")
        if isinstance(center, str):
            if center not in ['com']:
                raise ValueError(
                    "center must be a vector or the string 'com'.")
        elif np.size(center) != 3:
            raise ValueError(
                "center must either be a string or a 3-element vector.")

        # Convert to fixed reference frame
        if isinstance(coords, str):
            if coords=='fixed':
                if isinstance(center, str):
                    center = self.center_of_mass(coords='fixed')
            elif coords=='volume':
                translation = utils.volume_vector(translation, self.affine)
                rotation = utils.volume_vector(rotation, self.affine)
                if isinstance(center, str):
                    center = self.center_of_mass(coords='fixed')
                else:
                    center = utils.vol2fix(center, self.affine)
        else:
            translation = utils.volume_vector(translation, coords)
            rotation = utils.volume_vector(rotation, coords)
            if isinstance(center, str):
                center = self.center_of_mass(coords='fixed')
            else:
                center = utils.vol2fix(center, coords)
                
        # Get affine transformation
        return utils.affine_matrix(rotation, translation, stretch, center)


    def center_of_mass(self, coords='fixed'):
        """
        Center of mass.

        Args:
            coords (str or array, optional): Reference frame for the 
              coordinates of the returned vector, as a 4x4 affine array. 
              String options are shorthand notations: 'fixed' (patient 
              reference), and 'volume' (volume reference frame). Defaults 
              to 'fixed'.

        Returns:
            numpy.ndarray: 3-element vector pointing to the volume's center of 
            mass.
        """ 
        return utils.center_of_mass(self.values[:,:,:], self.affine, coords=coords)


    def reslice(self, affine=None, orient=None, rotation=None, center=None, 
                spacing=1.0, coords='fixed'):
        """Reslice the volume.

        Args:
            affine (array, optional): 4x4 affine array providing the affine 
              of the result. If this is not provided, the affine array is 
              constructed from the other arguments. Defaults to None.
            orient (str, optional): Orientation of the volume. The options are 
              'axial', 'sagittal', or 'coronal'. Alternatively the same options 
              can be provided referring to the orientation of the image planes: 
              'xy', 'yz' or 'zx'. If None is provided, the current 
              orientation of the volume is used. **Note** while 'xy' and 'yz' 
              mean exactly the same as 'axial' and 'sagittal', respectively, 
              'zx' differs from coronal in the orientation of the z-axis. While 
              'zx' forms a right-handed reference frame, 'coronal' is 
              left-handed following the common convention in medical imaging.
              Defaults to None.
            rotation (array, optional): 3-element array specifying the rotation 
              relative to *orient*, or relative to the current orientation 
              of the volume (if *orient* is None). Defaults to None.
            center (array, optional): 3-element array specifying the rotation 
              center of the new reference frame, in case a rotation is provided. 
              Defaults to None.
            spacing (float, optional): Pixel spacing in mm. Can be a 3D array or 
              a single scalar for isotropic spacing. Defaults to 1.0.
            coords (str or array, optional): Reference frame for the 
              coordinates of the rotation and center vector, as a 4x4 affine 
              array. String options are shorthand notations: 'fixed' (patient 
              reference frame), and 'volume' (volume reference frame). Defaults 
              to 'fixed'.

        Returns:
            Volume3D: resliced volume
        """
        if self.ndim > 3:
            raise ValueError("This function is not yet available for volumes "
                             "with more than 3 dimensions")
        if affine is None:
            
            # Convert to fixed coordinates
            if isinstance(coords, str):
                if coords=='volume':
                    if rotation is not None:
                        rotation = utils.volume_vector(rotation, self.affine)
                    if center is not None:
                        center = utils.vol2fix(center, self.affine)
            else:
                if rotation is not None:
                    rotation = utils.volume_vector(rotation, coords)
                if center is not None:
                    center = utils.vol2fix(center, coords)                

            # Determine affine
            if orient is None:
                transfo = utils.affine_matrix(
                    rotation=rotation, center=center, pixel_spacing=spacing)
                affine = transfo.dot(self.affine)
            else:
                affine = utils.make_affine(orient, rotation, center, spacing) 

        # Perform an active transform with the inverse
        transfo_inv = self.affine.dot(np.linalg.inv(affine))
        values, affine = mod_affine.affine_transform(
            self.values, self.affine, transfo_inv, reshape=True)
        
        # Transform the affine with the forward
        transfo = np.linalg.inv(transfo_inv)
        affine = transfo.dot(affine)

        return Volume3D(values, affine)

    
    def slice_like(self, v):
        """Slice the volume to the geometry of another volume

        Args:
            v (Volume3D or tuple): either a reference volume with 
                desired orientation and shape, or a tuple (shape, affine) 
                with a new shape and affine

        Returns:
            Volume3D: resliced volume
        """
        if self.ndim > 3:
            raise ValueError("This function is not yet available for volumes "
                             "with more than 3 dimensions")
        if isinstance(v, Volume3D):
            if v.ndim > 3:
                raise ValueError("This function is not yet available for volumes "
                                "with more than 3 dimensions")
            values, affine = mod_affine.affine_reslice(
                self.values, self.affine, 
                v.affine, output_shape=v.shape)
        else:
            values, affine = mod_affine.affine_reslice(
                self.values, self.affine, 
                v[1], output_shape=v[0])            
        return Volume3D(values, affine)

    
    def transform_affine_to(self, target, params, center=None, coords='fixed'):
        """Apply an affine transformation and reslice the result to the 
        geometry of a target volume.

        Args:
            target (vreg.Volume3D): target volume
            params (array-like): 9-element vector with translation vector,  
              rotation vector and stretch factors, in that order. 
            center (str or array-like, optional): center of rotation. If this 
              has value 'com' the rotation is performed 
              around the center of mass. Alternatively this can be a 
              3-element vector. Defaults to None (center = origin).
            coords (str or array, optional): Reference frame for the 
              coordinates of the translation, rotation and center vector, as 
              a 4x4 affine 
              array. String options are shorthand notations: 'fixed' (patient 
              reference frame), and 'volume' (volume reference frame). Defaults 
              to 'fixed'.

        Returns:
            vreg.Volume3D: transformed volume.
        """
        translation = params[:3]
        rotation = params[3:6]
        stretch = params[6:9]

        # Get affine transformation
        transform = self._affine_matrix(translation, rotation, center, 
                                        stretch, coords)

        # Apply affine transformation
        if self.ndim > 3:
            values_reshape = self.values.reshape(self.shape[:3] + (-1,))
            values_0 = mod_affine.affine_transform_and_reslice(
                values_reshape[...,0], self.affine, target.shape, target.affine, transform,
            )
            values = np.zeros(values_0.shape + (values.shape[-1],))
            values[...,0] = values_0
            for d in range(1, values.shape[-1]):
                values[...,d] = mod_affine.affine_transform_and_reslice(
                    self.values[...,d], self.affine, target.shape, target.affine, transform,
                )
            values = values.reshape(values_0.shape + self.shape[3:])
        else:
            values = mod_affine.affine_transform_and_reslice(
                self.values, self.affine, target.shape, target.affine, transform,
            )
        affine = target.affine.copy()
    
        # Return volume
        return Volume3D(values, affine)
    

    def transform_rigid_to(self, target, params, center=None, coords='fixed'):
        """Apply a rigid transformation and reslice the result to the 
        geometry of a target volume.

        Args:
            target (vreg.Volume3D): target volume
            params (array-like): 6-element vector with translation and 
              rotation vectors, in that order. 
            center (str or array-like): center of rotation. If this has value 
              'com' the rotation is performed around the 
              center of mass. Alternatively this can be a 3-element vector. 
              Defaults to None (center = origin).
            coords (str or array, optional): Reference frame for the 
              coordinates of the translation, rotation and center vector, as 
              a 4x4 affine 
              array. String options are shorthand notations: 'fixed' (patient 
              reference frame), and 'volume' (volume reference frame). Defaults 
              to 'fixed'.

        Returns:
            vreg.Volume3D: transformed volume.
        """

        stretch = np.ones(3)
        params = np.concatenate((params, stretch))
        return self.transform_affine_to(target, params, coords=coords, 
                                        center=center)
    

    def rotate_to(self, target, rotation, center=None, coords='fixed'):
        """Apply a rotation and reslice the result to the 
        geometry of a target volume.

        Args:
            target (vreg.Volume3D): target volume
            rotation (array-like): 3-element rotation vector in radians. 
              Defaults to None (no rotation).
            center (str or array-like): center of rotation. If this has value 
              'com' the rotation is performed around the 
              center of mass. Alternatively this can be a 3-element vector. 
              Defaults to None (center = origin).
            coords (str or array, optional): Reference frame for the 
              coordinates of the rotation and center vector, as a 4x4 affine 
              array. String options are shorthand notations: 'fixed' (patient 
              reference frame), and 'volume' (volume reference frame). Defaults 
              to 'fixed'.

        Returns:
            vreg.Volume3D: transformed volume.
        """
        translation = np.zeros(3)
        stretch = np.ones(3)
        params = np.concatenate((translation, rotation, stretch))
        return self.transform_affine_to(target, params, center=center, 
                                        coords=coords)


    def translate_to(self, target, translation, coords='fixed', dir='xyz'):
        """Apply a translation and reslice the result to the 
        geometry of a target volume.

        Args:
            target (vreg.Volume3D): target volume
            translation (array-like): translation vector (mm) with 1, 2 or 3 
              elements depending on the value of *dir*. 
            coords (str or array, optional): Reference frame for the 
              coordinates of the translation vector, as a 4x4 affine 
              array. String options are shorthand notations: 'fixed' (patient 
              reference frame), and 'volume' (volume reference frame). Defaults 
              to 'fixed'.
            dir (str, optional): Allowed directions of the translation. The 
              options are 'xyz' (3D translation), 'xy' (2D in-slice 
              translation) and 'z' (through-slice translation). Defaults to 
              'xyz'.

        Returns:
            vreg.Volume3D: transformed volume.
        """
        if dir=='xy':
            translation = np.concatenate((translation, [0]))
        elif dir=='z':
            translation = np.concatenate(([0,0], translation))
        rotation = np.zeros(3)
        stretch = np.ones(3)
        params = np.concatenate((translation, rotation, stretch))
        return self.transform_affine_to(target, params, coords=coords)
    

    def stretch_to(self, target, stretch):
        """Stretch and reslice to the geometry of a 
        target volume.

        Args:
            target (vreg.Volume3D): target volume
            stretch (array-like): 3-element stretch vector with strictly 
              positive dimensionless values (1 = no stretch).

        Returns:
            vreg.Volume3D: transformed volume.
        """
        translation = np.zeros(3)
        rotation = np.zeros(3)
        params = np.concatenate((translation, rotation, stretch))
        return self.transform_affine_to(target, params)
    

    def transform_affine(self, params, center=None, values=False, 
                         reshape=False, coords='fixed'):
        """Apply an affine transformation.

        Args:
            params (array-like): 9-element vector with translation vector,  
              rotation vector and stretch factors, in that order.
            center (str or array-like): center of rotation. If this has value 
              'com', the rotation is performed around the center of mass. 
              Alternatively this can be a 3-element vector. Defaults to None (
              center = origin).
            values (bool, optional): If set to True, the values are 
              transformed. Otherwise the affine is transformed. Defaults to 
              False.
            reshape (bool, optional): When values=True, reshape=False will 
              retain the shape and location of the volume. With reshape=True, 
              the volume will be reshaped to fit the transformed values. This 
              keyword is ignored when values=False. Defaults to False.
            coords (str or array, optional): Reference frame for the 
              coordinates of the translation, rotation and center vector, as 
              a 4x4 affine 
              array. String options are shorthand notations: 'fixed' (patient 
              reference frame), and 'volume' (volume reference frame). Defaults 
              to 'fixed'.

        Returns:
            vreg.Volume3D: transformed volume.
        """

        translation = params[:3]
        rotation = params[3:6]
        stretch = params[6:9]

        # Get affine transformation
        transform = self._affine_matrix(
            translation, rotation, center, stretch, coords
        )

        # Apply affine transformation
        if values:
            affine = self.affine.copy()
            if self.ndim > 3:
                values_reshape = self.values.reshape(self.shape[:3] + (-1,))
                values_0 = mod_affine.affine_transform(
                    values_reshape[...,0], self.affine, transform, reshape
                )
                values = np.zeros(values_0.shape + (values.shape[-1],))
                values[...,0] = values_0
                for d in range(1, values.shape[-1]):
                    values[...,d] = mod_affine.affine_transform(
                        values_reshape[...,d], self.affine, transform, reshape
                    )
                values = values.reshape(values_0.shape + self.shape[3:])
            else:
                values, _ = mod_affine.affine_transform(
                    self.values, self.affine, transform, reshape
                ) 
        else:
            values = self.values.copy()
            affine = transform.dot(self.affine)
        return Volume3D(values, affine, self.coords, self.dims, self.prec)
            

    def transform_rigid(self, params, values=False, reshape=False, 
                        center=None, coords='fixed'):
        """Apply a rigid transformation.

        Args:
            params (array-like): 6-element vector with translation and  
              rotation vectors, in that order. 
            values (bool, optional): If set to True, the values are 
              transformed. Otherwise the affine is transformed. Defaults to 
              False.
            reshape (bool, optional): When values=True, reshape=False will 
              retain the shape and location of the volume. With reshape=True, 
              the volume will be reshaped to fit the transformed values. This 
              keyword is ignored when values=False. Defaults to False.
            center (str or array-like): center of rotation. If this has value 
              'com', the rotation is performed around the center of mass. 
              Alternatively this can be a 3-element vector. Defaults to None (
              center = origin).
            coords (str or array, optional): Reference frame for the 
              coordinates of the translation, rotation and center vector, as 
              a 4x4 affine 
              array. String options are shorthand notations: 'fixed' (patient 
              reference frame), and 'volume' (volume reference frame). Defaults 
              to 'fixed'.

        Returns:
            vreg.Volume3D: transformed volume.
        """
        stretch = np.ones(3)
        params = np.concatenate((params, stretch))
        return self.transform_affine(params, center=center, values=values, 
                                     reshape=reshape, coords=coords)
        

    def rotate(self, rotation, center=None, values=False, reshape=False, 
               coords='fixed'):
        """Rotate the volume.

        Args:
            rotation (array-like): 3-element rotation vector in radians. 
            center (str or array-like): center of rotation. If this has value 
              'com', the rotation is performed around the center of mass. 
              Alternatively this can be a 3-element vector. Defaults to None (
              center = origin).
            values (bool, optional): If set to True, the values are 
              transformed. Otherwise the affine is transformed. Defaults to 
              False.
            reshape (bool, optional): When values=True, reshape=False will 
              retain the shape and location of the volume. With reshape=True, 
              the volume will be reshaped to fit the transformed values. This 
              keyword is ignored when values=False. Defaults to False.
            coords (str or array, optional): Reference frame for the 
              coordinates of the rotation and center vector, as a 4x4 affine 
              array. String options are shorthand notations: 'fixed' (patient 
              reference frame), and 'volume' (volume reference frame). Defaults 
              to 'fixed'.

        Returns:
            vreg.Volume3D: transformed volume.
        """
        translation = np.zeros(3)
        stretch = np.ones(3)
        params = np.concatenate((translation, rotation, stretch))
        return self.transform_affine(params, values=values, reshape=reshape, 
                                     center=center, coords=coords)


    def translate(self, translation, values=False, reshape=False, 
                  coords='fixed', dir='xyz'):
        """Translate the volume.

        Args:
            translation (array-like): translation vector (mm) with 1, 2 or 3 
              elements depending on the value of *dir*. 
            values (bool, optional): If set to True, the values are 
              transformed. Otherwise the affine is transformed. Defaults to 
              False.
            reshape (bool, optional): When values=True, reshape=False will 
              retain the shape and location of the volume. With reshape=True, 
              the volume will be reshaped to fit the transformed values. This 
              keyword is ignored when values=False. Defaults to False.
            coords (str or array, optional): Reference frame for the 
              coordinates of the translation vector, as a 4x4 affine array. 
              String options are shorthand notations: 'fixed' (patient 
              reference), and 'volume' (volume reference frame). Defaults 
              to 'fixed'.
            dir (str, optional): Allowed directions of the translation. The 
              options are 'xyz' (3D translation), 'xy' (2D in-slice 
              translation) and 'z' (through-slice translation). Defaults to 
              'xyz'.

        Returns:
            vreg.Volume3D: transformed volume.
        """
        if dir=='xy':
            translation = np.concatenate((translation, [0]))
        elif dir=='z':
            translation = np.concatenate(([0,0], translation))
        rotation = np.zeros(3)
        stretch = np.ones(3)
        params = np.concatenate((translation, rotation, stretch))
        return self.transform_affine(params, values=values, reshape=reshape, 
                                     coords=coords)

    def stretch(self, stretch, values=False, reshape=False):
        """Stretch the volume.

        Args:
            stretch (array-like): 3-element stretch vector or scalar with strictly 
              positive dimensionless values (1 = no stretch). 
            values (bool, optional): If set to True, the values are 
              transformed. Otherwise the affine is transformed. Defaults to 
              False.
            reshape (bool, optional): When values=True, reshape=False will 
              retain the shape and location of the volume. With reshape=True, 
              the volume will be reshaped to fit the transformed values. This 
              keyword is ignored when values=False. Defaults to False.

        Returns:
            vreg.Volume3D: transformed volume.
        """
        if np.isscalar(stretch):
            stretch = (stretch, stretch, stretch)
        translation = np.zeros(3)
        rotation = np.zeros(3)
        params = np.concatenate((translation, rotation, stretch))
        return self.transform_affine(params, values=values, reshape=reshape)


    def truncate(self, shape):
        """Truncate the volume to a smaller shape.

        Args:
            shape (array-like): 3-element array with the new shape. Each 
              dimension must be equal or smaller than the current shape.

        Returns:
            vreg.Volume3D: truncated volume.
        """
        if self.ndim > 3:
            return Volume3D(
                self.values[:shape[0], :shape[1], :shape[2], :],
                self.affine)
        return Volume3D(
            self.values[:shape[0], :shape[1], :shape[2]],
            self.affine)


        
def volume(values:np.ndarray, affine:np.ndarray=None, 
           coords:np.ndarray=None, dims:list=None, prec:int=None, 
           orient='axial', rotation=None, center=None, spacing=1.0, 
           pos=[0,0,0]):
    """Create a new volume from an array of values

    Args:
        values (array): 2D or 3D array of values.
        affine (array, optional): 4x4 affine array. If this is not provided, 
          the affine array is constructed from the other arguments. Defaults 
          to None.
        coords (list or tuple, optional): coordinates for volumes that 
          have more than 3 dimensions (non-spatial dimensions). If values 
          has N non-spatial dimensions, coords can be a list of N  
          coordinate arrays or a tuple of N 1D arrays. Values of the 
          coords array can have any type including strings or tuples.
        orient (str, optional): Orientation of the volume. The options are 
          'axial', 'sagittal', or 'coronal'. Alternatively the same options 
          can be provided referring to the orientation of the image planes: 
          'xy', 'yz' or 'zx'. **Note** while 'xy' and 'yz' mean exactly the 
          same as 'axial' and 'sagittal', respectively, 'zx' differs from 
          coronal in the orientation of the z-axis. While 'zx' forms a 
          right-handed reference frame, 'coronal' is left-handed following 
          the common convention in medical imaging.Defaults to 'axial.
        rotation (array, optional): 3-element array specifying a rotation 
          relative to *orient*. Defaults to None.
        center (array, optional): 3-element array specifying the rotation 
          center, in case a rotation is provided. Defaults to None.
        spacing (float, optional): Pixel spacing in mm. Can be a 3D array or 
          a single scalar for isotropic spacing. Defaults to 1.0.
        pos (list, optional): Position of the upper left-hand corner in mm. 
          Defaults to [0,0,0].

    Returns:
        vreg.Volume3D: volume with required orientation and position.
    """

    if affine is None:
        # Define orient=None and check that the user has defined an orientation
        affine = utils.make_affine(orient, rotation, center, spacing, pos) 
    if values.ndim==1:
        # This can only be done meaningfully if orient is specified
        values = np.expand_dims(values, -1)
    if values.ndim==2:
        # This can only be done meaningfully if orient is specified
        values = np.expand_dims(values, -1)
    if values.ndim>3:
        if coords is not None:
            if isinstance(coords, tuple):
                coords = [_list_to_array(c) for c in coords]
                coords = np.meshgrid(*coords, indexing='ij')
                coords = list(coords)  

    return Volume3D(values, affine, coords, dims, prec)



def _list_to_array(lst):
    """
    Helper function: convert a Python list to a NumPy array.
    
    - If all elements share the same dtype (e.g., all int, all float, all str),
      the returned array uses that dtype.
    - If elements have mixed types or contain sublists, an object array is returned.
    """
    if isinstance(lst, np.ndarray):
        return lst
    if lst == []:  # empty list
        return np.array([], dtype=object)

    # If any element is a list or types are mixed, use dtype=object
    element_types = {type(el) for el in lst}
    if any(isinstance(el, (list, tuple, np.ndarray)) for el in lst) or len(element_types) > 1:
        arr = np.empty(len(lst), dtype=object)
        arr[:] = lst
        return arr

    # Otherwise, let numpy infer a uniform dtype
    return np.array(lst)



def zeros(shape, affine=None, orient='axial', spacing=1.0, pos=[0,0,0], 
          **kwargs):
    """Return a new volume of given shape and affine, filled with zeros.

    Args:
        shape (int or tuple of ints): Shape of the new array, e.g., (2, 3) or 2.
        affine (array, optional): 4x4 affine array. If this is not provided, 
          the affine array is constructed from the other arguments. Defaults 
          to None.
        orient (str, optional): Orientation of the volume. The options are 
          'axial', 'sagittal', or 'coronal'. Alternatively the same options 
          can be provided referring to the orientation of the image planes: 
          'xy', 'yz' or 'zx'. **Note** while 'xy' and 'yz' mean exactly the 
          same as 'axial' and 'sagittal', respectively, 'zx' differs from 
          coronal in the orientation of the z-axis. While 'zx' forms a 
          right-handed reference frame, 'coronal' is left-handed following 
          the common convention in medical imaging. Defaults to 'axial'.
        spacing (float, optional): Pixel spacing in mm. Can be a 3D array or 
          a single scalar for isotropic spacing. Defaults to 1.0.
        pos (list, optional): Position of the upper left-hand corner in mm. 
          Defaults to [0,0,0].
        kwargs: Any keyword arguments accepted by `numpy.zeros`.

    Returns:
        Volume3D: vreg.Volume3D with zero values.
    """
    values = np.zeros(shape, **kwargs)
    return volume(values, affine, orient, spacing, pos)


def full(shape, fill_value, affine=None, orient='axial', spacing=1.0, 
         pos=[0,0,0], **kwargs):
    """Return a new volume of given shape and affine, filled with *fill_value*.

    Args:
        shape (int or tuple of ints): Shape of the new array, e.g., (2, 3) or 2.
        fill_value (float): value to fill the array.
        affine (array, optional): 4x4 affine array. If this is not provided, 
          the affine array is constructed from the other arguments. Defaults 
          to None.
        orient (str, optional): Orientation of the volume. The options are 
          'axial', 'sagittal', or 'coronal'. Alternatively the same options 
          can be provided referring to the orientation of the image planes: 
          'xy', 'yz' or 'zx'. **Note** while 'xy' and 'yz' mean exactly the 
          same as 'axial' and 'sagittal', respectively, 'zx' differs from 
          coronal in the orientation of the z-axis. While 'zx' forms a 
          right-handed reference frame, 'coronal' is left-handed following 
          the common convention in medical imaging. Defaults to 'axial'.
        spacing (float, optional): Pixel spacing in mm. Can be a 3D array or 
          a single scalar for isotropic spacing. Defaults to 1.0.
        pos (list, optional): Position of the upper left-hand corner in mm. 
          Defaults to [0,0,0].
        kwargs: Any keyword arguments accepted by `numpy.full`.

    Returns:
        Volume3D: vreg.Volume3D with *fill_value* values.
    """
    values = np.full(shape, fill_value, **kwargs)
    return volume(values, affine, orient, spacing, pos)


def zeros_like(v:Volume3D):
    """Return a new volume with same shape and affine as v, filled with zeros.

    Args:
        v (int or tuple of ints): Shape of the new array, e.g., (2, 3) or 2.

    Returns:
        Volume3D: vreg.Volume3D with zero values.
    """
    values = np.zeros(v.shape, dtype=v.values.dtype)
    return volume(values, v.affine.copy())


def full_like(v:Volume3D, fill_value):
    """Return a new volume with same shape and affine as v, filled with 
    *fill_value*.

    Args:
        v (int or tuple of ints): Shape of the new array, e.g., (2, 3) or 2.
        fill_value (float): value to fill the array.

    Returns:
        Volume3D: vreg.Volume3D with *fill_value* values.
    """
    values = np.full_like(v.values, fill_value)
    return volume(values, v.affine.copy())


def concatenate(vols, prec=None):
    """Join a sequence of volumes along x-, y-, or z-axis.

    Volumes can only be joined up if they have the same shape 
    (except along the axis of concatenation), the same orientation and 
    the same voxel size.

    Args:
        vols (sequence of volumes): Volumes to concatenate.
        prec (int, optional): precision to consider when comparing positions 
          and orientations of volumes. All differences are rounded off to this 
          digit before comparing them to zero. If this is not specified, 
          floating-point precision is used. Defaults to None.

    Returns:
        Volume3D: The concatenated volume.
    """
    if isinstance(vols, Volume3D):
        return vols
    
    # Check arguments
    if not np.iterable(vols):
        raise ValueError(
            "vreg.stack() requires an iterable as argument.")

    # Check that all volumes have the same shape and orientation
    if prec is None:
        mat = [v.affine[:3,:3].tolist() for v in vols]  
    else:
        mat = [np.around(v.affine[:3,:3], prec).tolist() for v in vols]
    mat = [x for i, x in enumerate(mat) if i==mat.index(x)]
    if len(mat) > 1:
        raise ValueError(
            "Volumes with different orientations or voxel sizes cannot be "
            "concatenated."
        )

    for axis in [0,1,2]:
        if _aligned_along_axis(vols, axis, prec):
            # Sort volumes according to position along concatenation axis
            vols = sorted(vols, key=lambda v: v.loc(axis))
            affine = vols[0].affine # use affine with smallest loc
            values = np.concatenate([v.values for v in vols], axis=axis)
            return Volume3D(values, affine)  
    raise ValueError(
        "Volumes cannot be concatenated. They are not aligned in "
        "any direction."
    )


def _aligned_along_axis(vols, axis, prec):
    axis_dir = vols[0].affine[:3,axis] / np.linalg.norm(vols[0].affine[:3,axis])
    for i in range(len(vols)-1):
        dz = vols[i+1].pos - vols[i].pos
        proj = np.abs(np.dot(dz, axis_dir))
        norm = np.linalg.norm(dz)
        diff = norm - proj
        if prec is not None:
            diff = np.around(diff, prec)
        if diff != 0:
            return False
    return True


# def _OLD_aligned_along_axis(vols, axis, prec):
#     mat = vols[0].affine[:3,:3]
#     pos = [v.affine[:3,3] for v in vols]
#     concat_vec = mat[:,axis]
#     for i, v in enumerate(vols[:-1]):
#         pos_next = pos[i] + concat_vec*v.shape[axis]
#         dist = np.linalg.norm(pos[i+1]-pos_next)
#         if prec is not None:
#             dist = np.around(dist, prec)
#         if dist > 0:
#             return False
#     return True





def stack(vols, axis=3, prec=None):
    """Stack a sequence volumes at the same location along a new time axis.

    Args:
        vols (sequence of volumes): Volumes to concatenate.
        axis (int, optional): The axis along which the volumes will be 
          stacked. This must be larger than 2.
        prec (int, optional): precision to consider when comparing positions 
          and orientations of volumes. All differences are rounded off to this 
          digit before comparing them to zero. If this is not specified, 
          floating-point precision is used. Defaults to None.

    Returns:
        Volume3D: The concatenated volume.
    """
    if isinstance(vols, Volume3D):
        return vols
    
    # Check arguments
    if not np.iterable(vols):
        raise ValueError(
            "vreg.stack() requires an iterable as argument.")
    if axis in [0,1,2]:
        raise ValueError(
            "Volumes cannot be stacked along a new space dimension. "
            "Consider using concatenate instead.")

    # Check that all volumes have the same shape and orientation
    if prec is None:
        mat = [v.affine[:3,:3].tolist() for v in vols]  
    else:
        mat = [np.around(v.affine[:3,:3], prec).tolist() for v in vols]
    mat = [x for i, x in enumerate(mat) if i==mat.index(x)]
    if len(mat) > 1:
        raise ValueError(
            "Volumes with different orientations or voxel sizes cannot be "
            "stacked."
        )
    
    # Check that all volumes are at the same position  
    for v in vols[1:]:
        dist = v.affine[:3,3] - vols[0].affine[:3,3]
        dist = np.linalg.norm(dist)
        if prec is not None:
            dist = np.around(dist, prec)
        if dist > 0:
            raise ValueError(
                "Volumes cannot be concatenated in a time direction. "
                "They are not all at the same position. Set move=True "
                "if you want to allow them to move to the correct "
                "position.")

    # Determine concatenation and return new volume
    affine = vols[0].affine
    values = np.stack([v.values for v in vols], axis=axis)
    return Volume3D(values, affine)  


def join(vols:np.ndarray): #concat->stack, join as a generalized stack, stack-> concat
    """Join multiple volumes into a single volume.

    This is the opposite operation to split

    Args:
        vols (array of volumes): Volumes to join. 

    Returns:
        Volume3D: A single volume.
    """

    # Concatenate in space
    shape = vols.shape
    vols = vols.reshape((shape[0],-1))
    vols_stack = []
    for k in range(vols.shape[1]):
        vstack = concatenate(vols[:,k], prec=2) # TODO use internal precision 
        vols_stack.append(vstack)
    if len(shape) == 1:
        return vols_stack[0]
    else:
        vols = np.asarray(vols_stack).reshape(shape[1:])
        
    # Stack along the other dimensions
    axis = 3
    while True:
        shape = vols.shape
        vols = vols.reshape((shape[0],-1))
        vols_stack = []
        for k in range(vols.shape[1]):
            vstack = stack(vols[:,k], axis=axis, prec=3) 
            vols_stack.append(vstack)
        if len(shape) == 1:
            return vols_stack[0]
        else:
            vols = np.asarray(vols_stack).reshape(shape[1:])
            axis += 1


def mean(vol:Volume3D, axis=None): # replicates numpy function
    vals = np.mean(vol.values, axis=axis)
    if axis is None:
        vals = np.full(vol.shape, vals)
        return Volume3D(vals, vol.affine.copy())
    elif axis < 3:
        vals = np.expand_dims(vals, axis=axis)
        return Volume3D(vals, vol.affine.copy())
    if axis >= 3:
        dims = [d for i, d in enumerate(vol.dims) if i != axis-3]
        coords = [np.take(c, indices=0, axis=axis-3) for c in vol.coords]
        if dims==[]:
            dims = None
            coords = None
        else:
            # THIS NEEDS DEBUGGING
            inds = [i for i in range(coords.shape[0]) if i != axis]
            coords = [np.take(c, indices=inds, axis=0) for c in vol.coords]
        mean_vol = Volume3D(vals, vol.affine.copy(), coords, dims)
    return mean_vol

# def asarray(vol):
    
#     if isinstance(vol, Volume3D):
#         return vol.values
    
#     vol = np.asarray(vol)

#     # Check that all affines are the same
#     uniq = np.unique([v.affine for v in vol.reshape(-1)])
#     if uniq.size>1:
#         raise ValueError("Cannot extract an array from volumes that have "
#                          "different affines.")
    
#     # Check that all shapes are the same
#     uniq = np.unique([v.shape for v in vol.reshape(-1)])
#     if uniq.size>1:
#         raise ValueError("Cannot extract an array from volumes that have "
#                          "different shapes.")

#     # Build n-dimensional array
#     shape = vol.shape
#     arr = np.stack([v.values for v in vol.reshape(-1)], axis=-1)
#     return arr.reshape(arr.shape[:3] + shape)


# def asvolume(arr:np.ndarray, affine):
#     if arr.ndim <= 3:
#         return volume(arr, affine)
#     shape = arr.shape[3:]
#     arr = arr.reshape((arr.shape[:3], -1))
#     vol = [volume(arr[...,k], affine) for k in range(arr.shape[-1])]
#     return np.asarray(vol).reshape(shape)


# def _take_view(arr, i, axis):
#     # alternative to np.take() which returns a view
#     # Create a tuple of slice(None) (which selects all elements) for each axis
#     slc = [slice(None)] * arr.ndim
#     # Replace the desired axis with the scalar index
#     slc[axis] = i
#     return arr[tuple(slc)]  # Returns a view

def _take_view_keepdims(arr, i, axis):
    slc = [slice(None)] * arr.ndim
    slc[axis] = slice(i, i + 1)
    return arr[tuple(slc)]

# def _take_keepdims(arr, i, axis):
#     slc = [slice(None)] * arr.ndim
#     slc[axis] = slice(i, i + 1)
#     return arr[tuple(slc)].copy()










import os
import numpy as np

import vreg

from vreg.data import DATASETS


def test_fetch():
    mask = vreg.fetch('left_kidney')
    assert 1504 == np.sum(mask.values[:,:,2])
    vreg.fetch(clear_cache=True)
    vreg.fetch(download_all=True)
    vreg.fetch(download_all=True)

# def convert():
#     tmp = os.path.join(os.getcwd(), 'tmp')
#     os.makedirs(tmp, exist_ok=True)
#     for data in ['T1', 'T2star']:
#         vol = vreg.fetch(data)
#         for i, v, in enumerate(vol):
#             vreg.write_nifti(v, f"{os.path.join(tmp, data)}_{str(i).zfill(2)}")

if __name__ == '__main__':

    #test_fetch()
    vreg.fetch(clear_cache=True)
    #convert()


import numpy as np
from skimage.draw import ellipsoid

import vreg.utils
import vreg.mod_affine

#############################
# Generate test data
#############################


def generate(structure='ellipsoid', shape=None, affine=None, markers=True):
    
    # Default shape
    if shape is None:
        shape = (256, 256, 40) 

    # Default affine
    if affine is None:
        pixel_spacing = np.array([1.5, 1.5, 5]) # mm
        translation = np.array([0, 0, 0]) # mm
        rotation_angle = -0.2 * (np.pi/2) # radians
        rotation_axis = [1,0,0]
        rotation = rotation_angle * np.array(rotation_axis)/np.linalg.norm(rotation_axis)
        affine = vreg.utils.affine_matrix(rotation=rotation, translation=translation, pixel_spacing=pixel_spacing)

    _, _ , pixel_spacing = vreg.utils.affine_components(affine) 
    data = np.zeros(shape, dtype=np.float32)  

    if markers:
        # Insert cube markers at corners
        marker_width = 20 # marker width in mm
        w = np.around(np.divide(np.array([marker_width]*3), pixel_spacing))
        w = w.astype(np.int16) 
        data[0:w[0],0:w[1],0:w[2]] = 1
        data[-w[0]:,0:w[1],0:w[2]] = 1
        data[0:w[0],-w[1]:,0:w[2]] = 1
        data[-w[0]:,-w[1]:,0:w[2]] = 1
        data[0:w[0],0:w[1],-w[2]:] = 1
        data[-w[0]:,0:w[1],-w[2]:] = 1
        data[0:w[0],-w[1]:,-w[2]:] = 1
        data[-w[0]:,-w[1]:,-w[2]:] = 1

    if structure == 'ellipsoid':
        half_length = (20, 30, 40) # mm
        ellip = ellipsoid(half_length[0], half_length[1], half_length[2], spacing=pixel_spacing, levelset=False)
        d = ellip.shape
        p = [30, 30, 10]
        data[p[0]:p[0]+d[0], p[1]:p[1]+d[1], p[2]:p[2]+d[2]] = ellip
        return data, affine
    
    elif structure == 'double ellipsoid':
        half_length1 = np.array([10, 20, 30]) # mm
        half_length2 = np.array([5, 10, 15]) # mm
        pos = np.array([150, 50, 0]) # mm
        ellip1 = ellipsoid(half_length1[0], half_length1[1], half_length1[2], spacing=pixel_spacing, levelset=False)
        ellip2 = ellipsoid(half_length2[0], half_length2[1], half_length2[2], spacing=pixel_spacing, levelset=False)
        ellip1 = ellip1.astype(np.int16)
        ellip2 = ellip2.astype(np.int16)

        p = np.around(np.divide(pos, pixel_spacing)).astype(np.int16)
        d = ellip1.shape
        data[p[0]:p[0]+d[0], p[1]:p[1]+d[1], p[2]:p[2]+d[2]] = ellip1

        p += np.around([d[0], d[1]/4, d[2]/2]).astype(np.int16)
        d = ellip2.shape
        data[p[0]:p[0]+d[0], p[1]:p[1]+d[1], p[2]:p[2]+d[2]] = ellip2

        return data, affine

    elif structure == 'triple ellipsoid':
        half_length1 = np.array([10, 20, 30]) # mm
        half_length2 = np.array([5, 10, 15]) # mm
        p1 = np.array([150, 50, 10]) # mm
        p2 = np.array([170, 70, 20]) # mm
        p3 = np.array([150, 150, 10]) # mm

        ellip1 = ellipsoid(half_length1[0], half_length1[1], half_length1[2], spacing=pixel_spacing, levelset=False)
        ellip2 = ellipsoid(half_length2[0], half_length2[1], half_length2[2], spacing=pixel_spacing, levelset=False)
        ellip1 = ellip1.astype(np.int16)
        ellip2 = ellip2.astype(np.int16)

        p = np.around(np.divide(p1, pixel_spacing)).astype(np.int16)
        d = ellip1.shape
        data[p[0]:p[0]+d[0], p[1]:p[1]+d[1], p[2]:p[2]+d[2]] = ellip1
        
        p = np.around(np.divide(p2, pixel_spacing)).astype(np.int16)
        d = ellip2.shape
        data[p[0]:p[0]+d[0], p[1]:p[1]+d[1], p[2]:p[2]+d[2]] = ellip2

        p = np.around(np.divide(p3, pixel_spacing)).astype(np.int16)
        d = ellip1.shape
        data[p[0]:p[0]+d[0], p[1]:p[1]+d[1], p[2]:p[2]+d[2]] = ellip1

        return data, affine 


def generate_plot_data_1():

    # Define geometry of input data
    pixel_spacing = np.array([2.0, 2.0, 10.0]) # mm
    input_shape = np.array([100, 100, 10], dtype=np.int16)
    translation = np.array([0, 0, 0]) # mm
    rotation_angle = 0.0 * (np.pi/2) # radians
    rotation_axis = [1,0,0]
    rotation = rotation_angle * np.array(rotation_axis)/np.linalg.norm(rotation_axis)
    input_affine = vreg.utils.affine_matrix(rotation=rotation, translation=translation, pixel_spacing=pixel_spacing)
    
    # Generate ground truth data
    input_data, input_affine = generate('triple ellipsoid', shape=input_shape, affine=input_affine, markers=True)

    return input_data, input_affine


def generate_reslice_data_1():

    # Downsample
    # Reslice high-res volume to lower resolution.
    # Values are chosen so that the field of view stays the same.

    # Define geometry of input data
    matrix = np.array([400, 300, 120])  
    pixel_spacing = np.array([1.0, 1.0, 1.0]) # mm
    translation = np.array([0, 0, 0]) # mm
    rotation_angle = -0.0 * (np.pi/2) # radians
    rotation_axis = [1,0,0]
    rotation = rotation_angle * np.array(rotation_axis)/np.linalg.norm(rotation_axis)

    # Define geometry of output data
    output_pixel_spacing = np.array([1.25, 5.0, 10.0]) # mm
    output_shape = None # retain field of view

    # Generate data
    input_affine = vreg.utils.affine_matrix(rotation=rotation, translation=translation, pixel_spacing=pixel_spacing)
    output_affine = vreg.utils.affine_matrix(rotation=rotation, translation=translation, pixel_spacing=output_pixel_spacing)
    input_data, input_affine = generate('triple ellipsoid', shape=matrix, affine=input_affine, markers=True)

    return input_data, input_affine, output_affine, output_shape


def generate_reslice_data_2():

    # Upsample
    # Reslice low-res volume to higher resolution.
    # Values are chosen so that the field of view stays the same.

    # Define geometry of input data
    matrix = np.array([320, 60, 8])  
    pixel_spacing = np.array([1.25, 5.0, 15.0]) # mm
    translation = np.array([0, 0, 0]) # mm
    rotation_angle = -0.0 * (np.pi/2) # radians
    rotation_axis = [1,0,0]
    rotation = rotation_angle * np.array(rotation_axis)/np.linalg.norm(rotation_axis)

    # Define geometry of output data
    output_pixel_spacing = np.array([1.0, 1.0, 1.0]) # mm
    output_shape = None # retain field of view

    # Generate data
    input_affine = vreg.utils.affine_matrix(rotation=rotation, translation=translation, pixel_spacing=pixel_spacing)
    output_affine = vreg.utils.affine_matrix(rotation=rotation, translation=translation, pixel_spacing=output_pixel_spacing)
    input_data, input_affine = generate('triple ellipsoid', shape=matrix, affine=input_affine, markers=False)

    return input_data, input_affine, output_affine, output_shape


def generate_reslice_data_3():

    # resample to lower resolution with a
    # 90 degree rotation around x + translation along y

    # Define source data
    matrix = np.array([400, 300, 120]) 
    pixel_spacing = np.array([1.0, 1.0, 1.0]) # mm
    translation = np.array([0, 0, 0]) # mm
    rotation_angle = 0 * (np.pi/2) # radians
    rotation_axis = [1,0,0]

    # Generate source data
    rotation = rotation_angle * np.array(rotation_axis)/np.linalg.norm(rotation_axis)
    input_affine = vreg.utils.affine_matrix(rotation=rotation, translation=translation, pixel_spacing=pixel_spacing)
    input_data, input_affine = generate('triple ellipsoid', shape=matrix, affine=input_affine)

    # Define geometry of new slab
    pixel_spacing = np.array([1.25, 1.25, 10.0]) # mm
    translation = np.array([0, 120, 0]) # mm
    rotation_angle = 1.0 * (np.pi/2) # radians
    rotation_axis = [1,0,0]
    
    # Reslice current slab to geometry of new slab
    rotation = rotation_angle * np.array(rotation_axis)/np.linalg.norm(rotation_axis)
    output_affine = vreg.utils.affine_matrix(rotation=rotation, translation=translation, pixel_spacing=pixel_spacing)
    output_shape = None
    return input_data, input_affine, output_affine, output_shape


def generate_reslice_data_4():

    # Rotation at low resolution

    # Define geometry of input data
    input_shape = np.array([40, 40, 20], dtype=np.int16)
    pixel_spacing = np.array([6, 6, 6.0])       # mm
    translation = np.array([0, 0, 0]) # mm
    rotation_angle = 0.0 * (np.pi/2) # radians
    rotation_axis = [1,0,0]
    rotation = rotation_angle * np.array(rotation_axis)/np.linalg.norm(rotation_axis)
    input_affine = vreg.utils.affine_matrix(rotation=rotation, translation=translation, pixel_spacing=pixel_spacing)

    # Define output affine
    #pixel_spacing = np.array([0.5, 0.5, 0.5])       # mm
    translation = np.array([0, 0, 0])     # mm
    rotation_angle = 0.15 * (np.pi/2)    # radians
    #translation = np.array([0, 0, 30]) # mm
    rotation_axis = [1,0,0]
    rotation = rotation_angle * np.array(rotation_axis)/np.linalg.norm(rotation_axis)
    output_affine = vreg.utils.affine_matrix(rotation=rotation, translation=translation, pixel_spacing=pixel_spacing)
    
    # Generate input data data
    input_data, input_affine = generate('triple ellipsoid', shape=input_shape, affine=input_affine, markers=True)

    return input_data, input_affine, output_affine, None


def generate_reslice_data_5():

    # Reslice an object with its own affine

    # Define geometry of input data
    input_size = np.array([400, 300, 120])   # mm
    input_shape = np.array([400, 300, 120], dtype=np.int16)
    translation = np.array([0, 0, 0]) # mm
    rotation_angle = 0.0 * (np.pi/2) # radians
    rotation_axis = [1,0,0]
    rotation = rotation_angle * np.array(rotation_axis)/np.linalg.norm(rotation_axis)
    pixel_spacing = np.divide(input_size, input_shape)
    input_affine = vreg.utils.affine_matrix(rotation=rotation, translation=translation, pixel_spacing=pixel_spacing)

    # Define output affine
    pixel_spacing = np.array([1.6, 2.6, 7.5])       # mm
    translation = np.array([100, 0, 0])     # mm
    rotation_angle = 0.1 * (np.pi/2)    # radians
    rotation_axis = [1,0,0]
    rotation = rotation_angle * np.array(rotation_axis)/np.linalg.norm(rotation_axis)
    output_affine = vreg.utils.affine_matrix(rotation=rotation, translation=translation, pixel_spacing=pixel_spacing)
    
    # Generate input data data
    input_data, input_affine = generate('triple ellipsoid', shape=input_shape, affine=input_affine, markers=False)

    # Reslice to output affine keeping the same field of view
    output_data, output_affine = vreg.mod_affine.affine_reslice(input_data, input_affine, output_affine)

    return output_data, output_affine, output_affine, None


def generate_reslice_data_6():

    # 1-pixel thick - does not work yet!!

    # Define geometry of input data
    pixel_spacing = np.array([1.25, 1.25, 5.0]) # mm
    input_shape = np.array([300, 200, 20], dtype=np.int16)
    translation = np.array([0, 0, 0]) # mm
    rotation_angle = 0.0 * (np.pi/2) # radians
    rotation_axis = [1,0,0]
    rotation = rotation_angle * np.array(rotation_axis)/np.linalg.norm(rotation_axis)
    input_affine = vreg.utils.affine_matrix(rotation=rotation, translation=translation, pixel_spacing=pixel_spacing)

    # Define geometry of output data
    pixel_spacing = np.array([1.25, 1.25, 10.0]) # mm
    output_shape = np.array([300, 250, 1], dtype=np.int16)
    translation = np.array([0, 0, 5]) # mm
    rotation_angle = 0.2 * (np.pi/2) # radians
    rotation_axis = [1,0,0]
    rotation = rotation_angle * np.array(rotation_axis)/np.linalg.norm(rotation_axis)
    output_affine = vreg.utils.affine_matrix(rotation=rotation, translation=translation, pixel_spacing=pixel_spacing)
    
    # Generate ground truth data
    input_data, input_affine = generate('triple ellipsoid', shape=input_shape, affine=input_affine, markers=True)

    return input_data, input_affine, output_affine, output_shape


def generate_translated_data_1():

    # Define geometry of input data
    pixel_spacing = np.array([1.25, 1.25, 5.0]) # mm
    translation = np.array([0, 0, 0]) # mm
    rotation_angle = 0.5 * (np.pi/2) # radians
    rotation_axis = [1,0,0]
    rotation = rotation_angle * np.array(rotation_axis)/np.linalg.norm(rotation_axis)
    input_affine = vreg.utils.affine_matrix(rotation=rotation, translation=translation, pixel_spacing=pixel_spacing)
    input_shape = np.array([300, 200, 25])  

    # Define affine transformation
    translation = np.array([10, -10, 10]) # mm
    
    # Define geometry of output data (exactly equal to translated volume)
    transformation = vreg.utils.affine_matrix(translation=translation)
    output_shape, output_affine = vreg.utils.affine_output_geometry(input_shape, input_affine, transformation)

    # Generate ground truth data
    input_data, input_affine = generate('triple ellipsoid', shape=input_shape, affine=input_affine, markers=False)
    output_data = vreg.mod_affine.translate(input_data, input_affine, output_shape, output_affine, translation)

    return input_data, input_affine, output_data, output_affine, translation


def generate_translated_data_2():

    # Model for 3D to 2D registration

    # Define geometry of input data
    pixel_spacing = np.array([1.0, 1.0, 1.0]) # mm
    translation = np.array([0, 0, 0]) # mm
    rotation_angle = 0.0 * (np.pi/2) # radians
    rotation_axis = [1,0,0]
    rotation = rotation_angle * np.array(rotation_axis)/np.linalg.norm(rotation_axis)
    input_affine = vreg.utils.affine_matrix(rotation=rotation, translation=translation, pixel_spacing=pixel_spacing)
    input_shape = np.array([300, 200, 100]) 

    # Define affine transformation
    active_translation = np.array([10, -10, 10]) # mm

    # Define geometry of output data
    output_shape = np.array([150, 200, 1])  
    pixel_spacing = np.array([1.25, 1.25, 7.5]) # mm
    #translation = np.array([100, 0, 50]) # mm
    translation = np.array([100, 0, 25]) # mm
    rotation_angle = 0.1 * (np.pi/2) # radians
    rotation_axis = [1,0,0]
    rotation = rotation_angle * np.array(rotation_axis)/np.linalg.norm(rotation_axis)
    output_affine = vreg.utils.affine_matrix(rotation=rotation, translation=translation, pixel_spacing=pixel_spacing)

    # Generate ground truth data
    input_data, input_affine = generate('triple ellipsoid', shape=input_shape, affine=input_affine, markers=False)
    output_data = vreg.mod_affine.translate(input_data, input_affine, output_shape, output_affine, active_translation)

    return input_data, input_affine, output_data, output_affine, active_translation


def generate_translated_data_3():

    # Model for 2D to 3D registration
    # Same as 2 but input and output reversed

    # Define geometry of input data
    pixel_spacing = np.array([1.0, 1.0, 1.0]) # mm
    translation = np.array([0, 0, 0]) # mm
    rotation_angle = 0.0 * (np.pi/2) # radians
    rotation_axis = [1,0,0]
    rotation = rotation_angle * np.array(rotation_axis)/np.linalg.norm(rotation_axis)
    input_affine = vreg.utils.affine_matrix(rotation=rotation, translation=translation, pixel_spacing=pixel_spacing)
    input_shape = np.array([300, 200, 100]) 

    # Define affine transformation
    active_translation = np.array([10, -10, 10]) # mm

    # Define geometry of output data
    output_shape = np.array([150, 200, 1])   # mm
    pixel_spacing = np.array([1.25, 1.25, 7.5]) # mm
    # translation = np.array([100, 0, 50]) # mm
    translation = np.array([100, 0, 25]) # mm
    rotation_angle = 0.1 * (np.pi/2) # radians
    rotation_axis = [1,0,0]
    rotation = rotation_angle * np.array(rotation_axis)/np.linalg.norm(rotation_axis)
    output_affine = vreg.utils.affine_matrix(rotation=rotation, translation=translation, pixel_spacing=pixel_spacing)

    # Generate ground truth data
    input_data, input_affine = generate('triple ellipsoid', shape=input_shape, affine=input_affine, markers=False)
    output_data = vreg.mod_affine.translate(input_data, input_affine, output_shape, output_affine, active_translation)

    return output_data, output_affine, input_data, input_affine, -active_translation



from vreg import optimize
from vreg.optimize import *

from vreg import utils
from vreg.utils import *

from vreg import mod_affine
from vreg.mod_affine import *

from vreg import metrics
from vreg.metrics import *

from vreg import mod_freeform
from vreg.mod_freeform import *

from vreg import fake
from vreg.fake import *

from vreg import vol
from vreg.vol import *

# from vreg import plot
# from vreg.plot import *

from vreg import data
from vreg.data import *

from vreg import rw
from vreg.rw import *


# from vreg import scipy_least_squares
# from vreg.scipy_least_squares import *


import numpy as np
import scipy.ndimage as ndi

from scipy.spatial.transform import Rotation
from scipy.interpolate import griddata 

def affine_resolution(shape, spacing):
    """Smallest detectable rotation, translation and stretching of a volume with given shape and resolution."""

    translation_res = spacing
    # Geometry-based rotation resolution needs some more thinking - this produces too small values
    # rot_res_x = min([spacing[1],spacing[2]])/max([shape[1],shape[2]])
    # rot_res_y = min([spacing[2],spacing[0]])/max([shape[2],shape[0]])
    # rot_res_z = min([spacing[0],spacing[1]])/max([shape[0],shape[1]])
    # rot_res = np.array([rot_res_x, rot_res_y, rot_res_z])
    rot_res = np.array([np.pi/180, np.pi/180, np.pi/180])
    scaling_res = np.array([0.01, 0.01, 0.01])
    return rot_res, translation_res, scaling_res


def volume_coordinates(shape, position=[0,0,0]):

    # data are defined at the middle of the voxels - use p+0.5 offset.
    xo, yo, zo = np.meshgrid(
        np.arange(position[0]+0.5, position[0]+0.5+shape[0]),
        np.arange(position[1]+0.5, position[1]+0.5+shape[1]),
        np.arange(position[2]+0.5, position[2]+0.5+shape[2]),
        indexing = 'ij')
    return np.column_stack((xo.ravel(), yo.ravel(), zo.ravel()))


def fill_gaps(data, loc, mask=None):
    # Fill gaps in data by interpolation
    # data is an array with values
    # loc is a mask array defining where to interpolate (0=interpolate, 1=value)
    # mask is a mask array to restrict the interpolation to certain regions.

    x, y, z = np.indices(data.shape)

    # Get locations and values of non-zero pixels
    i = loc>0.5
    if mask is not None:
        i = i & (mask==1)
    points = np.column_stack([x[i], y[i], z[i]])
    values = data[i]

    # Interpolate using griddata
    k = loc<0.5
    if mask is not None:
        k = k & (mask==1)
    filled = data.copy()
    filled[k] = griddata(points, values, (x[k], y[k], z[k]), method='linear', fill_value=0)

    if mask is not None:
        filled *= mask

    return filled


def rotation_displacement(rotation, center):

    rot = Rotation.from_rotvec(rotation)
    center_rot = rot.apply(center)
    return center_rot-center


def envelope(d, affine, decimals=None):

    corners, _ = parallellepid(np.array(d), affine)

    x0 = np.amin(corners[:,0])
    x1 = np.amax(corners[:,0])
    y0 = np.amin(corners[:,1])
    y1 = np.amax(corners[:,1])
    z0 = np.amin(corners[:,2])
    z1 = np.amax(corners[:,2])

    dx = x1-x0
    dy = y1-y0
    dz = z1-z0
    if decimals is not None:
        dx = np.round(dx, decimals)
        dy = np.round(dy, decimals)
        dz = np.round(dz, decimals)
    nx = np.ceil(dx).astype(np.int16)
    ny = np.ceil(dy).astype(np.int16)
    nz = np.ceil(dz).astype(np.int16)

    output_shape = (nx, ny, nz)
    output_pos = [x0, y0, z0]

    return output_shape, output_pos

def bounding_box_geom(array:np.ndarray, affine:np.ndarray, margin:float=0)->tuple:

    pixel_spacing = np.linalg.norm(affine[:3, :3], axis=0)
    xmargin = np.around(margin/pixel_spacing[0]).astype(np.int16)
    ymargin = np.around(margin/pixel_spacing[1]).astype(np.int16)
    zmargin = np.around(margin/pixel_spacing[2]).astype(np.int16)

    # Find shape and location of the box in array coordinates and get array.
    x, y, z = np.where(array != 0)
    x0, x1 = np.amin(x)-xmargin, np.amax(x)+xmargin
    y0, y1 = np.amin(y)-ymargin, np.amax(y)+ymargin
    z0, z1 = np.amin(z)-zmargin, np.amax(z)+zmargin
    x0, x1 = np.amax([0, x0]), np.amin([x1, array.shape[0]-1])
    y0, y1 = np.amax([0, y0]), np.amin([y1, array.shape[1]-1])
    z0, z1 = np.amax([0, z0]), np.amin([z1, array.shape[2]-1])
    nx = 1 + np.ceil(x1-x0).astype(np.int16)
    ny = 1 + np.ceil(y1-y0).astype(np.int16)
    nz = 1 + np.ceil(z1-z0).astype(np.int16)
    return x0, y0, z0, nx, ny, nz

def bounding_box_apply(array, affine, x0, y0, z0, nx, ny, nz):

    box_array = array[x0:x0+nx, y0:y0+ny, z0:z0+nz]

    # Get the corner in absolute coordinates and offset the affine.
    nd = 3
    matrix = affine[:nd,:nd]
    offset = affine[:nd, nd]
    r0 = np.array([x0,y0,z0])
    r0 = np.dot(r0, matrix.T) + offset
    box_affine = affine.copy()
    box_affine[:nd, nd] = r0

    return box_array, box_affine


def bounding_box(array:np.ndarray, affine:np.ndarray, margin:float=0)->tuple:

    # pixel_spacing = np.linalg.norm(affine[:3, :3], axis=0)
    # xmargin = np.around(margin/pixel_spacing[0]).astype(np.int16)
    # ymargin = np.around(margin/pixel_spacing[1]).astype(np.int16)
    # zmargin = np.around(margin/pixel_spacing[2]).astype(np.int16)

    # # Find shape and location of the box in array coordinates and get array.
    # x, y, z = np.where(array != 0)
    # x0, x1 = np.amin(x)-xmargin, np.amax(x)+xmargin
    # y0, y1 = np.amin(y)-ymargin, np.amax(y)+ymargin
    # z0, z1 = np.amin(z)-zmargin, np.amax(z)+zmargin
    # x0, x1 = np.amax([0, x0]), np.amin([x1, array.shape[0]-1])
    # y0, y1 = np.amax([0, y0]), np.amin([y1, array.shape[1]-1])
    # z0, z1 = np.amax([0, z0]), np.amin([z1, array.shape[2]-1])
    # nx = 1 + np.ceil(x1-x0).astype(np.int16)
    # ny = 1 + np.ceil(y1-y0).astype(np.int16)
    # nz = 1 + np.ceil(z1-z0).astype(np.int16)

    x0, y0, z0, nx, ny, nz = bounding_box_geom(array, affine, margin)
    return bounding_box_apply(array, affine, x0, y0, z0, nx, ny, nz)
    # box_array = array[x0:x0+nx, y0:y0+ny, z0:z0+nz]

    # # Get the corner in absolute coordinates and offset the affine.
    # nd = 3
    # matrix = affine[:nd,:nd]
    # offset = affine[:nd, nd]
    # r0 = np.array([x0,y0,z0])
    # r0 = np.dot(r0, matrix.T) + offset
    # box_affine = affine.copy()
    # box_affine[:nd, nd] = r0

    # return box_array, box_affine


def affine_output_geometry(input_shape, input_affine, transformation):
        
    # Determine output shape and position
    affine_transformed = transformation.dot(input_affine)
    forward = np.linalg.inv(input_affine).dot(affine_transformed) # Ai T A
    output_shape, output_pos = envelope(input_shape, forward, decimals=6)

    # Determine output affine by shifting affine to the output position
    nd = input_affine.shape[0]-1
    matrix = input_affine[:nd,:nd]
    offset = input_affine[:nd, nd]
    output_affine = input_affine.copy()
    output_affine[:nd, nd] = np.dot(matrix, output_pos) + offset

    return output_shape, output_affine


def multislice_to_singleslice_affine(affine_ms, slice_thickness):
    # In a multi-slice affine, the z-spacing is the slice spacing (distance between slice centers)
    # In a single-slice affine, the z-spacing is the slice thickness
    affine_ss = affine_ms.copy()
    affine_ss[:3,2] *= slice_thickness/np.linalg.norm(affine_ss[:3,2])
    return affine_ss

def affine_slice(affine, z, slice_thickness=None):
    # Get the slice and its affine
    affine_z = affine.copy()
    affine_z[:3,3] += z*affine[:3,2]
    # Set the slice spacing to equal the slice thickness.
    # Note: both are equal for 3D array but different for 2D multislice
    # This caters for the convention in dbdicom and should become obsolete
    if slice_thickness is not None:
        slice_spacing = np.linalg.norm(affine[:3,2])
        if np.isscalar(slice_thickness):
            affine_z[:3,2] *= slice_thickness/slice_spacing
        else:
            affine_z[:3,2] *= slice_thickness[z]/slice_spacing
    return affine_z


def extract_slice(array, affine, z, slice_thickness=None):
    array_z = array[:,:,z]
    affine_z = affine_slice(affine, z, slice_thickness=slice_thickness)
    return array_z, affine_z

def center_of_mass(volume, affine, coords='fixed'):

    com = ndi.center_of_mass(volume)
    if isinstance(coords, str):
        if coords == 'volume':
            return com
        if coords == 'fixed':
            return vol2fix(com, affine)
    else:
        return vol2vol(com, affine, coords)

    # nd = volume.ndim
    # matrix = affine[:nd,:nd]
    # offset = affine[:nd, nd]
    # return np.dot(matrix, com) + offset

def vol2vol(vec, affine_in, affine_out):
    affine = np.linalg.inv(affine_out).dot(affine_in)
    matrix = affine[:3,:3]
    offset = affine[:3, 3]
    return np.dot(matrix, vec) + offset
    # vec = vol2fix(vec, affine_in)
    # vec = fix2vol(vec, affine_out)
    # return vec

def vol2fix(vec, affine):
    matrix = affine[:3,:3]
    offset = affine[:3, 3]
    return np.dot(matrix, vec) + offset

def fix2vol(vec, affine):
    affine = np.linalg.inv(affine)
    matrix = affine[:3,:3]
    offset = affine[:3, 3]
    return np.dot(matrix, vec) + offset

def volume_vector(vec, affine):
    row_cosine = affine[:3,0]/np.linalg.norm(affine[:3,0])
    column_cosine = affine[:3,1]/np.linalg.norm(affine[:3,1])
    slice_cosine = affine[:3,2]/np.linalg.norm(affine[:3,2])
    # Replaced as incorrect for stretched volume?
    # slice_cosine = np.cross(row_cosine, column_cosine)
    return vec[0]*row_cosine + vec[1]*column_cosine + vec[2]*slice_cosine

def inslice_vector(vec, affine):
    row_cosine = affine[:3,0]/np.linalg.norm(affine[:3,0])
    column_cosine = affine[:3,1]/np.linalg.norm(affine[:3,1])
    return vec[0]*row_cosine + vec[1]*column_cosine

def through_slice_vector(vec, affine):
    # Replaced as incorrect for stretched volume?
    # row_cosine = affine[:3,0]/np.linalg.norm(affine[:3,0])
    # column_cosine = affine[:3,1]/np.linalg.norm(affine[:3,1])
    # slice_cosine = np.cross(row_cosine, column_cosine)
    slice_cosine = affine[:3,2]/np.linalg.norm(affine[:3,2])
    if np.isscalar(vec):
        return vec*slice_cosine
    else:
        return vec[0]*slice_cosine


def to_3d(array):
    # Ensures that the data are 3D
    # Obsolete as Volume3D enforces 3D
    if array.ndim == 2: 
        return np.expand_dims(array, axis=-1)
    else:
        return array
    

def make_affine(orient='axial', rotation=None, center=None, spacing=1.0, 
                pos=[0,0,0]):

    # Check data types
    if orient not in ['axial','coronal','sagittal', 'xy', 'yz', 'zx']:
        raise ValueError(
            str(orient) + " is not a valid orientation. Possible "
            "orientations are 'axial', 'coronal' and 'sagittal', "
            "or also 'xy', 'yz' and 'zx'. ")

    if rotation is not None:
        rotation = np.array(rotation)
        if rotation.size != 3:
            raise ValueError("rotation must be a 3-element array.")
        
    if center is not None:
        center = np.array(center)
        if center.size != 3:
            raise ValueError("center must be a 3-element array.")
        
    if np.isscalar(spacing):
        spacing = 3*[spacing]
    spacing = np.array(spacing)
    if spacing.size != 3:
        raise ValueError("spacing must be a 3-element array.")
    
    # Set up default orientation
    if orient in ['axial', 'xy']:
        affine = np.array([[1, 0, 0, 0],
                           [0, 1, 0, 0],
                           [0, 0, 1, 0],
                           [0, 0, 0, 1]], dtype=np.float32)
    elif orient in ['sagittal', 'yz']:
        affine = np.array([[ 0, 0, 1, 0],
                           [-1, 0, 0, 0],
                           [ 0,-1, 0, 0],
                           [ 0, 0, 0, 1]], dtype=np.float32)
    elif orient == 'coronal':
        affine = np.array([[1, 0, 0, 0],
                           [0, 0,-1, 0],
                           [0,-1, 0, 0],
                           [0, 0, 0, 1]], dtype=np.float32)
    elif orient == 'zx':
        affine = np.array([[1, 0, 0, 0],
                           [0, 0, 1, 0],
                           [0,-1, 0, 0],
                           [0, 0, 0, 1]], dtype=np.float32)
        
    # Add spacing
    for i in [0,1,2]:
        affine[:,i] *= spacing[i]

    # Rotate default orientation
    if rotation is not None:  
        transfo = affine_matrix(rotation=rotation, center=center) 
        affine = transfo.dot(affine)
  
    # Shift
    affine[:3,3] = np.array(pos)

    return affine
    

def affine_matrix(rotation=None, translation=None, pixel_spacing=None, center=None):

    if np.isscalar(pixel_spacing):
        pixel_spacing = 3*[pixel_spacing]

    nd = 3
    matrix = np.eye(1+nd)

    if rotation is not None:
        rot = Rotation.from_rotvec(rotation)
        matrix[:nd,:nd] = rot.as_matrix()

        # Shift to rotate around center
        if center is not None:
            center_rot = rot.apply(center)
            offset = center_rot-center
            matrix[:nd, nd] -= offset

    if translation is not None:
        matrix[:nd, nd] += translation

    if pixel_spacing is not None:
        for c in range(nd):
            matrix[:nd, c] *= pixel_spacing[c]

    return matrix


def affine_components(matrix):
    """Extract rotation, translation and pixel spacing vector from affine matrix"""

    nd = matrix.shape[0]-1
    translation = matrix[:nd, nd].copy()
    rotation_matrix = matrix[:nd, :nd].copy()
    pixel_spacing = np.linalg.norm(matrix[:nd, :nd], axis=0)
    for c in range(nd):
        rotation_matrix[:nd, c] /= pixel_spacing[c]
    rot = Rotation.from_matrix(rotation_matrix)
    rotation = rot.as_rotvec()
    return rotation, translation, pixel_spacing


def surface_coordinates(shape):

    # data are defined at the edge of volume - extend shape with 1.
    xo, yo, zo = np.meshgrid(
        np.arange(1.0 + shape[0]),
        np.arange(1.0 + shape[1]),
        np.arange(1.0 + shape[2]),
        indexing = 'ij')
    return np.column_stack((xo.ravel(), yo.ravel(), zo.ravel()))


def extend_border(r, shape):

    # Shift with half a voxel because data are defined at voxel centers
    r -= 0.5

    # Set coordinates at 0.5 pixel from the borders equal to the borders
    x0, x1 = 0, shape[0]-1
    y0, y1 = 0, shape[1]-1
    z0, z1 = 0, shape[2]-1

    r[np.where(np.logical_and(x0-0.5 <= r[:,0], r[:,0] <= x0)), 0] = x0
    r[np.where(np.logical_and(x1+0.5 >= r[:,0], r[:,0] >= x1)), 0] = x1
    r[np.where(np.logical_and(y0-0.5 <= r[:,1], r[:,1] <= y0)), 1] = y0
    r[np.where(np.logical_and(y1+0.5 >= r[:,1], r[:,1] >= y1)), 1] = y1
    r[np.where(np.logical_and(z0-0.5 <= r[:,2], r[:,2] <= z0)), 2] = z0
    r[np.where(np.logical_and(z1+0.5 >= r[:,2], r[:,2] >= z1)), 2] = z1

    return r


def parallellepid(L, affine=None):

    c = np.array([0,0,0])
    x = np.array([1,0,0])*L[0]
    y = np.array([0,1,0])*L[1]
    z = np.array([0,0,1])*L[2]

    # mesh points
    vertices = np.array(
        [   c, 
            c+x, 
            c+x+z, 
            c+z, 
            c+y, 
            c+y+x, 
            c+y+x+z, 
            c+y+z,
        ]
    )
    
    if affine is not None:
        nd = 3
        matrix = affine[:nd,:nd]
        offset = affine[:nd, nd]
        vertices = np.dot(vertices, matrix.T) + offset

    # mesh faces
    faces = np.hstack(
        [
            [4, 0, 1, 2, 3], #right
            [4, 4, 5, 6, 7], #left
            [4, 0, 1, 5, 4], #bottom
            [4, 2, 3, 7, 6], #top
            [4, 0, 3, 7, 4], #front
            [4, 1, 2, 6, 5], #back
        ]
    )

    return vertices, faces


import numpy as np
import vreg



def test_coords():

    shape = (4,5,2,3,9)
    vals = np.zeros(shape)
    affine = np.eye(4)

    # Well-formatted coordinates - mixed type
    X = np.full((3,9), 'A')
    Y = np.ones((3,9))
    vol = vreg.volume(vals, affine, [X, Y])
    assert vol.coords[0][-1,-1] == 'A'
    assert vol.coords[1][-1,-1] == 1

    # Well-formatted coordinates - tuple of arrays
    X = np.full(3, 'A')
    Y = np.ones(9)
    vol = vreg.volume(vals, affine, (X, Y))
    assert vol.coords[0][-1,-1] == 'A'
    assert vol.coords[1][-1,-1] == 1

    # Well-formatted coordinates - tuple of lists
    X = ['A'] * 3
    Y = [1] * 9
    vol = vreg.volume(vals, affine, (X, Y))
    assert vol.coords[0][-1,-1] == 'A'
    assert vol.coords[1][-1,-1] == 1

    # Well-formatted coordinates - tuple of lists
    X = ['A', 'B', 'C']
    Y = np.arange(9)
    vol = vreg.volume(vals, affine, (X, Y))
    assert vol.coords[0][-1,-1] == 'C'
    assert vol.coords[1][-1,-1] == 8

    # Well-formatted coordinates - objectarray
    X = np.array(['A', 'B', 'C'], dtype=object)
    Y =  np.arange(9)
    vol = vreg.volume(vals, affine, (X, Y))
    assert vol.coords[0][-1,-1] == 'C'
    assert vol.coords[1][-1,-1] == 8

    # Well-formatted coordinates - object array from mixed type list
    X = np.array(['A', ['A','B'], ['A','B','C']], dtype=object)
    Y = np.arange(9)
    vol = vreg.volume(vals, affine, (X, Y))
    assert vol.coords[0][-1,-1] == ['A', 'B', 'C']
    assert vol.coords[1][-1,-1] == 8

    # Well-formatted coordinates - object array from list of lists
    X = np.array([['A'], ['A','B'], ['A','B','C']], dtype=object)
    Y = np.arange(9)
    vol = vreg.volume(vals, affine, (X, Y))
    assert vol.coords[0][-1,-1] == ['A', 'B', 'C']
    assert vol.coords[1][-1,-1] == 8

    # Well-formatted coordinates - list of lists
    X = [['A'], ['A','B'], ['A','B','C']]
    Y = np.arange(9)
    vol = vreg.volume(vals, affine, (X, Y))
    assert vol.coords[0][-1,-1] == ['A', 'B', 'C']
    assert vol.coords[1][-1,-1] == 8

    # Well-formatted coordinates - list of lists
    X = [['A'], ['A','B'], ['A','B','C']]
    Y = ['X'] * 9
    vol = vreg.volume(vals, affine, (X, Y))
    assert vol.coords[0][-1,-1] == ['A', 'B', 'C']
    assert vol.coords[1][-1,-1] == 'X'

    # Well-formatted coordinates - mixed type list
    X = ['A', ['A','B'], ['A','B','C']]
    Y = np.arange(9)
    vol = vreg.volume(vals, affine, (X, Y))
    assert vol.coords[0][0,-1] == 'A'
    assert vol.coords[0][-1,-1] == ['A', 'B', 'C']
    assert vol.coords[1][-1,-1] == 8

    # Well-formatted coordinates - mixed type list
    X = ['A', ['A','B'], 2]
    Y = np.arange(9)
    vol = vreg.volume(vals, affine, (X, Y))
    assert vol.coords[0][0,-1] == 'A'
    assert vol.coords[0][-1,-1] == 2
    assert vol.coords[1][-1,-1] == 8

    # Well-formatted coordinates - mixed type array
    X = np.array(['A', True, 2], dtype=object)
    Y = np.arange(9)
    vol = vreg.volume(vals, affine, (X, Y))
    assert vol.coords[0][0,-1] == 'A'
    assert vol.coords[0][1,-1] == True
    assert vol.coords[0][2,-1] == 2
    assert vol.coords[1][-1,-1] == 8

    # Well-formatted coordinates - mixed type list
    X = ['A', True, 2]
    Y = np.arange(9)
    vol = vreg.volume(vals, affine, (X, Y))
    assert vol.coords[0][0,-1] == 'A'
    assert vol.coords[0][1,-1] == True
    assert vol.coords[0][2,-1] == 2
    assert vol.coords[1][-1,-1] == 8



if __name__ == '__main__':

    test_coords()

    print('All creation tests passed!!!')


import numpy as np
import scipy.optimize
from tqdm import tqdm
# import dask
# from dask.diagnostics import ProgressBar

from vreg.scipy_least_squares import least_squares

# @dask.delayed
# def _compute_cost(cost_function, x, i, *args):
#     parameters = np.array([xp.ravel()[i] for xp in x])
#     return cost_function(parameters, *args)

# @dask.delayed
# def _opt_params(x, cost):
#     i = np.argmin(cost)
#     return np.array([xp.ravel()[i] for xp in x])

# def _minimize_brute(cost_function, args=None, grid=None, 
#                    desc='Performing brute-force optimization', progress=True):
#     x = [np.linspace(p[0], p[1], p[2]) for p in grid]
#     x = np.meshgrid(*tuple(x), indexing='ij')
#     cost = []
#     for i in range(x[0].size):
#     #for i in tqdm(range(x[0].size), desc=desc, disable=not progress):
#         ci = _compute_cost(cost_function, x, i, *args)
#         cost.append(ci)
#     opt = _opt_params(x, cost)
#     #with ProgressBar():
#         #opt = opt.compute(scheduler='single-threaded')
#     opt = opt.compute(scheduler='single-threaded')
#     return opt

def minimize(goodness_of_alignment, parameters, args=(), method='LS', **kwargs):
    "General minimization function including scipy and custom methods"

    if method == 'GD':
        return minimize_gd(goodness_of_alignment, parameters, args=args, **kwargs)
    elif method=='brute':
        return minimize_brute(goodness_of_alignment, args=args, **kwargs)
    elif method=='brute-LS':
        return minimize_brute_ls(goodness_of_alignment, args=args, **kwargs)
    elif method=='ibrute':
        return minimize_iterative_brute(goodness_of_alignment, args=args, **kwargs)
    elif method=='LS':
        res = least_squares(goodness_of_alignment, parameters, args=args, **kwargs)
        return res.x
    elif method == 'min':
        res = scipy.optimize.minimize(goodness_of_alignment, parameters, args=args, **kwargs)
        return res.x
    else:
        raise ValueError(
            "Optimization method " + str(method) + " is not recognized."
            "The options are 'GD', 'brute', 'ibrute', 'LS' or 'min'.")

def minimize_brute_ls(cost_function, args=None, grid=None, **kwargs):
    params = minimize_brute(cost_function, args=args, grid=grid)
    res = least_squares(cost_function, params, args=args, **kwargs)
    return res.x

def minimize_brute(
        cost_function, args=None, grid=None, 
        desc='Performing brute-force optimization', progress=False, 
        callback=None):
    #grid = [[start, stop, num], [start, stop, num], ...]
    x = [np.linspace(p[0], p[1], p[2]) for p in grid]
    x = np.meshgrid(*tuple(x), indexing='ij')
    for i in tqdm(range(x[0].size), desc=desc, disable=not progress):
        if callback is not None:
            callback(100*(i+1)/x[0].size)
        parameters = np.array([xp.ravel()[i] for xp in x])
        cost = cost_function(parameters, *args)
        if i==0:
            minimum = cost
            par_opt = parameters
        elif cost<minimum:
            minimum = cost
            par_opt = parameters
    return par_opt


def minimize_iterative_brute(goodness_of_alignment, args=None, num=3, nit=5, bounds=None):
    #options = {'num':3, 'nit':5, 'bounds':[[start, stop], [start, stop], ...]}
    grid = [[b[0], b[1], num] for b in bounds]
    parameters = minimize_brute(goodness_of_alignment, args=args, grid=grid)
    for _ in range(nit-1):
        dist = [(g[1]-g[0])/(num-1) for g in grid]
        grid = [[p-dist[i], p+dist[i], num] for i, p in enumerate(parameters)]
        parameters = minimize_brute(goodness_of_alignment, args=args, grid=grid)
    return parameters


def minimize_gd(
        cost_function, parameters, args=None, callback=None, max_iter=100, 
        bounds=None, gradient_step=None, tolerance=0.1, scale_down=5.0, 
        scale_up=1.5, stepsize_max=1000.0):

    # Set default values for global options
    if gradient_step is None:
        gradient_step = np.ones(parameters.shape)

    stepsize = 1.0 # initial stepsize
    n_iter = 0
    cost = cost_function(parameters, *args)
    while True:
        n_iter+=1
        print('iteration: ', n_iter)
        grad = gradient(cost_function, parameters, cost, gradient_step, bounds, *args)
        parameters, stepsize, cost = line_search(
            cost_function, grad, parameters, stepsize, cost, bounds, *args, 
            tolerance=tolerance, scale_down=scale_down, scale_up=scale_up, 
            stepsize_max=stepsize_max)
        if callback is not None:
            callback(parameters)
        if cost == 0:
            return parameters
        if stepsize == 0:
            return parameters
        if n_iter == max_iter:
            return parameters


def gradient(cost_function, parameters, f0, step, bounds, *args):
    grad = np.empty(parameters.shape)
    for i in range(parameters.size):
        c = np.unravel_index(i, parameters.shape)
        pc = parameters[c]
        sc = step[c]
        parameters[c] = pc+sc
        parameters = project_on(parameters, bounds, coord=c)
        fp = cost_function(parameters, *args)
        parameters[c] = pc-sc
        parameters = project_on(parameters, bounds, coord=c)
        fn = cost_function(parameters, *args)
        parameters[c] = pc
        grad[c] = (fp-fn)/2
        #grad[i] = (fp-fn)/(2*step[i])
        #grad[i] = stats.linregress([-step[i],0,step[i]], [fn,f0,fp]).slope

    # Normalize the gradient
    grad_norm = np.linalg.norm(grad)
    if grad_norm == 0:
        return grad 
    grad /= grad_norm
    grad = np.multiply(step, grad)

    return grad


def project_on(par, bounds, coord=None):
    if bounds is None:
        return par
    if len(bounds) != len(par):
        msg = 'Parameter and bounds must have the same length'
        raise ValueError(msg)
    if coord is not None:   # project only that index
        pc = par[coord]
        bc = bounds[coord]
        if pc <= bc[0]:
            pc = bc[0]
        if pc >= bc[1]:
            pc = bc[1]
    else:   # project all indices
        for i in range(par.size):
            c = np.unravel_index(i, par.shape)
            pc = par[c]
            bc = bounds[c]
            if pc <= bc[0]:
                par[c] = bc[0]
            if pc >= bc[1]:
                par[c] = bc[1]
    return par


def line_search(
        cost_function, grad, p0, stepsize0, f0, bounds, *args, 
        tolerance=0.1, scale_down=5.0, scale_up=1.5, stepsize_max=1000.0):

    # Initialize stepsize to current optimal stepsize
    stepsize_try = stepsize0 / scale_down
    p_init = p0.copy()

    # backtrack in big steps until reduction in cost
    while True:

        # Take a step and evaluate the cost
        p_try = p_init - stepsize_try*grad 
        p_try = project_on(p_try, bounds)
        f_try = cost_function(p_try, *args)

        print('cost: ', f_try, ' stepsize: ', stepsize_try, ' par: ', p_try)

        # If a reduction in cost is found, move on to the next part
        if f_try < f0:
            break

        # Otherwise reduce the stepsize and try again
        else:
            stepsize_try /= scale_down

        # If the stepsize has been reduced below the resolution without reducing the cost,
        # then the initial values were at the minimum (stepsize=0).
        if stepsize_try < tolerance: 
            return p0, 0, f0 # converged
        
    if stepsize_try < tolerance: 
        return p_try, 0, f_try # converged

    # If a reduction in cost has been found, then refine it 
    # by moving forward in babysteps until the cost increases again.
    while True:
        
        # Update the current optimum
        stepsize0 = stepsize_try
        f0 = f_try
        p0 = p_try

        # Take a baby step and evaluate the cost
        stepsize_try *= scale_up
        p_try = p_init - stepsize_try*grad
        p_try = project_on(p_try, bounds)
        f_try = cost_function(p_try, *args)

        print('cost: ', f_try, ' stepsize: ', stepsize_try, ' par: ', p_try)

        # If the cost has increased then a minimum was found
        if f_try >= f0:
            return p0, stepsize0, f0

        # emergency stop
        if stepsize_try > stepsize_max:
            msg = 'Line search failed to find a minimum'
            raise ValueError(msg) 